# -*- coding: utf-8 -*-
#
# Copyright (C) 2008 OpenSourceXpress Ltd. (http://www.ossxp.com)
# Author: Jiang Xin
# Contact: http://www.ossxp.com
#          http://www.worldhello.net
#          http://moinmo.in/JiangXin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import cgi

from paste.urlparser import PkgResourcesParser
from pylons.middleware import error_document_template
from webhelpers.html.builder import literal
from pylons.i18n import _, ungettext, N_

from pysvnmanager.lib.base import BaseController

class ErrorController(BaseController):
    """Generates error documents as and when they are required.

    The ErrorDocuments middleware forwards to ErrorController when error
    related status codes are returned from the application.

    This behaviour can be altered by changing the parameters to the
    ErrorDocuments middleware in your config/middleware.py file.

    """
    def document(self):
        """Render the error document"""
        request = self._py_object.request
        resp = request.environ.get('pylons.original_response')
        content = literal(resp.body) or cgi.escape(request.GET.get('message', ''))
        page = error_document_template % \
            dict(prefix=request.environ.get('SCRIPT_NAME', ''),
                 code=cgi.escape(request.GET.get('code', str(resp.status_int))),
                 message=content)
        return page

    def img(self, id):
        """Serve Pylons' stock images"""
        return self._serve_file('/'.join(['media/img', id]))

    def style(self, id):
        """Serve Pylons' stock stylesheets"""
        return self._serve_file('/'.join(['media/style', id]))

    def _serve_file(self, path):
        """Call Paste's FileApp (a WSGI application) to serve the file
        at the specified path
        """
        request = self._py_object.request
        request.environ['PATH_INFO'] = '/%s' % path
        return PkgResourcesParser('pylons', 'pylons')(request.environ, self.start_response)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2008 OpenSourceXpress Ltd. (http://www.ossxp.com)
# Author: Jiang Xin
# Contact: http://www.ossxp.com
#          http://www.worldhello.net
#          http://moinmo.in/JiangXin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import os
import sys
sys.path.insert(0,os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))

from pysvnmanager.tests import *
from pysvnmanager import model
from pysvnmanager.model import rcsbackup as rcs
import StringIO
import time

class TestRcsBackup(TestController):
    wcfile  = "%s/%s" % (os.path.dirname(os.path.abspath(__file__)), 'rcstest.txt')
    wcpath = os.path.dirname(os.path.abspath(wcfile))
    if os.path.isdir(wcpath+'/RCS'):
        rcsfile = wcpath+'/RCS/'+os.path.basename(wcfile)+',v'
    else:
        rcsfile = wcfile+',v'

    #def __init__(self, *args):
    #    super(TestController, self).__init__(*args)
 
    def setUp(self):
        if os.access(self.wcfile, os.R_OK):
            os.remove(self.wcfile)
        if os.access(self.rcsfile, os.R_OK):
            os.remove(self.rcsfile)

    def tearDown(self):
        if os.access(self.wcfile, os.R_OK):
            os.remove(self.wcfile)
        if os.access(self.rcsfile, os.R_OK):
            os.remove(self.rcsfile)

    def writefile(self, rev=None):
        if not rev:
            rev = self.get_revision()+1
        elif not isinstance(rev, int):
            rev = int(rev)

        f = open(self.wcfile, 'w')
        f.write('RCS working copy file\n')
        f.write('='*20 + '\n')
        f.write('Revision: %d\n' % rev)
        f.write('Date: %s\n' % time.strftime('%F'))
        f.close()

    def get_revision(self):
        rev = 0
        if os.access(self.wcfile, os.F_OK):
            f = open(self.wcfile, 'r')
            for line in f:
                line = line.strip()
                if line.startswith('Revision: '):
                    rev = int(line.split(':',1)[1])
                    break
        return rev

    def testBackup(self):
        # Backup test. (rcs file not exist yet)
        self.writefile()
        assert self.get_revision() == 1, self.get_revision()
        assert os.access(self.wcfile, os.R_OK)
        assert not os.access(self.rcsfile, os.R_OK)
        rcs.backup(self.wcfile)
        assert os.access(self.wcfile, os.R_OK)
        assert os.path.exists(self.rcsfile)
        assert os.access(self.rcsfile, os.R_OK), self.rcsfile

        # Backup test. (rcs exist already)
        self.writefile()
        assert os.access(self.rcsfile, os.R_OK)
        rcs.backup(self.wcfile)

        # exception test
        self.assertRaises(Exception, rcs.backup, "")

    def testRestore(self):
        rcs.restore("")
        
        # new file, backup to r1.1
        self.writefile()
        rcs.backup(self.wcfile)

        # restore from top rev: 1.1
        os.remove(self.wcfile)
        assert not os.access(self.wcfile, os.R_OK)
        rcs.restore(self.wcfile)
        assert self.get_revision() == 1, self.get_revision()

        # new file, backup to 1.2
        self.writefile()
        rcs.backup(self.wcfile)

        # backup from top rev: 1.2, and overwrite wcfile
        self.writefile()
        assert self.get_revision() == 3, self.get_revision()
        rcs.restore(self.wcfile)
        assert self.get_revision() == 2, self.get_revision()

        # backup from old rev: 1.1, and overwrite wcfile
        rcs.restore(self.wcfile, '1.1')
        assert self.get_revision() == 1, self.get_revision()

        # new file, backup to 1.3
        self.writefile(5)
        rcs.backup(self.wcfile)
        assert self.get_revision() == 5, self.get_revision()

        # restore from top rev: 1.3
        os.remove(self.wcfile)
        assert not os.access(self.wcfile, os.R_OK)
        rcs.restore(self.wcfile)
        assert self.get_revision() == 5, self.get_revision()

    def testRevision(self):
        self.writefile()
        assert os.access(self.wcfile, os.R_OK)
        pass

    def testLogs(self):
        rcslog = rcs.RcsLog(self.wcfile)
        rcslog.log_per_page = 10
        for i in range(1,rcslog.log_per_page+1):
            # new file, backup to r1.1
            self.writefile(i)
            rcs.backup(self.wcfile, comment="Test no. %d" % i, user="User1")
            rcslog.reload()
            assert rcslog.total == i, rcslog.total
            assert rcslog.total_page == 1, rcslog.total_page
        
        rcslog.reload()
        assert "rcstest.txt,v" in rcslog.rcsfile, rcslog.rcsfile
        assert rcslog.head=='1.10', rcslog.head
        assert rcslog.total==10, rcslog.total

        for i in range(11,15):
            # new file, backup to r1.1
            self.writefile(i)
            rcs.backup(self.wcfile, comment="第 %d 次提交测试。" % i, user="蒋鑫")

        assert rcslog.head=='1.10', rcslog.head
        rcslog.reload()
        assert rcslog.head=='1.14', rcslog.head
        assert rcslog.total==14, rcslog.total
        
        logs = rcslog.get_logs()
        assert len(logs)==14, logs
        assert logs[3]['revision'] == u'1.4', logs[3]['revision']
        assert logs[3]['author'] == u'User1', logs[3]['author']
        assert logs[3]['log'] == u'Test no. 4', logs[3]['log']
        assert logs[13]['revision'] == u'1.14', logs[13]['revision']
        assert logs[13]['author'] == u'蒋鑫', logs[13]['author'].encode('utf-8')
        assert logs[13]['log'] == u'第 14 次提交测试。', logs[13]['log'].encode('utf-8')
        
        logs = rcslog.get_logs('1.9','1.12')
        assert len(logs)==4, logs
        assert logs[1]['revision'] == u'1.10', logs[1]['revision']
        assert logs[1]['author'] == u'User1', logs[1]['author']
        assert logs[1]['log'] == u'Test no. 10', logs[1]['log']
        assert logs[2]['revision'] == u'1.11', logs[2]['revision']
        assert logs[2]['author'] == u'蒋鑫', logs[2]['author'].encode('utf-8')
        assert logs[2]['log'] == u'第 11 次提交测试。', logs[2]['log'].encode('utf-8')


        logs = rcslog.get_logs('','1.12')
        assert len(logs)==12, logs
        assert logs[10]['revision'] == u'1.11', logs[10]['revision']

        logs = rcslog.get_logs('1.12','')
        assert len(logs)==3, len(logs)
        assert logs[1]['revision'] == u'1.13', logs[1]['revision']

        logs = rcslog.get_logs('1.7','1.11', '1.14')
        assert len(logs)==6, len(logs)
        assert logs[4]['revision'] == u'1.11', logs[1]['revision']
        assert logs[5]['revision'] == u'1.14', logs[1]['revision']
        
        assert rcslog.total_page == 2, rcslog.total_page
        rcslog.log_per_page=0
        assert rcslog.log_per_page>1, rcslog.log_per_page
        rcslog.log_per_page=5
        assert rcslog.log_per_page==5, rcslog.log_per_page
        assert rcslog.total_page == 3, rcslog.total_page
        logs = rcslog.get_page_logs(1)
        logs2= rcslog.get_page_logs(0)
        assert logs == logs2
        assert [x['revision'] for x in logs] == \
                ['1.9', '1.10', '1.11', '1.12', '1.13', '1.14'], \
                [x['revision'] for x in logs]

        logs = rcslog.get_page_logs(2)
        assert [x['revision'] for x in logs] == \
                ['1.5', '1.6', '1.7', '1.8', '1.9', '1.14'], \
                [x['revision'] for x in logs]

        logs = rcslog.get_page_logs(3)
        logs2= rcslog.get_page_logs(30)
        assert logs == logs2
        assert [x['revision'] for x in logs] == \
                ['1.1', '1.2', '1.3', '1.4', '1.5', '1.14'], \
                [x['revision'] for x in logs]

        buff = rcslog.differ('1.2','1.5')
        assert "-Revision: 2\n+Revision: 5" in buff, buff
        buff = rcslog.differ('1.13')
        assert "-Revision: 13\n+Revision: 14" in buff, buff
        
        buff = rcslog.cat()
        assert "Revision: 14\n" in buff, buff

        buff = rcslog.cat('1.8')
        assert "Revision: 8\n" in buff, buff
        
    def testLogsNone(self):
        rcslog = rcs.RcsLog(self.wcfile)
        assert rcslog.rcsfile == "", rcslog.rcsfile
        assert rcslog.total == 0, rcslog.total
        assert rcslog.get_page_logs(1) == [], rcslog.get_page_logs(1)
        assert rcslog.cat()==""
        
if __name__ == '__main__': 
    import unittest
    unittest.main()




## -*- coding: utf-8 -*-
#
# Copyright (C) 2008 OpenSourceXpress Ltd. (http://www.ossxp.com)
# Author: Jiang Xin
# Contact: http://www.ossxp.com
#          http://www.worldhello.net
#          http://moinmo.in/JiangXin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

from pysvnmanager.tests import *
from pysvnmanager.controllers import role

class TestRoleController(TestController):

    def test_index(self):
        # Test redirect to login pange
        res = self.app.get(url(controller='role', action='index'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/login'), res.location

        # Login as common user
        self.login('nobody')
        res = self.app.get(url(controller='role', action='index'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/security/failed'), res.location
        
        # Repos admin(not root admin) can access role controller, but with disabled button.
        self.login('admin2')
        res = self.app.get(url(controller='role', action='index'))
        assert res.status == "200 OK", res.status
        assert """
    <input type="button" name="save_btn"   value='Save'  onClick="do_save(this.form)" DISABLED>
    <input type="button" name="delete_btn" value='Delete' onClick="do_delete(this.form)" DISABLED>
    <input type="button" name="cancel_btn" value='Cancel' onClick="role_changed()" DISABLED>""" in res.body, res.body[-300:]

        # Login as superuser
        self.login('root')
        res = self.app.get(url(controller='role', action='index'))
        assert res.status == "200 OK", res.status
        assert """<input type="button" name="save_btn"   value='Save'""" in res.body, res.body

    def test_get_role_info(self):
        # authn test
        res = self.app.get(url(controller='role', action='get_role_info'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/login'), res.location

        # authz test
        self.login('nobody')
        res = self.app.get(url(controller='role', action='get_role_info'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/security/failed'), res.location
        
        # Login as superuser
        self.login('root')
        params = {'role':'',}
        res = self.app.get(url(controller='role', action='get_role_info'), params)
        assert res.status == "200 OK", res.status
        assert """id[0]="...";name[0]="Please choose...";
id[1]="@admin";name[1]="Group:admin";
id[2]="@all";name[2]="Group:all";
id[3]="@biz";name[3]="Group:biz";
id[4]="@dev";name[4]="Group:dev";
id[5]="@group1";name[5]="Group:group1";
id[6]="@group2";name[6]="Group:group2";
id[7]="@group3";name[7]="Group:group3";
id[8]="@office";name[8]="Group:office";
id[9]="@tech";name[9]="Group:tech";
id[10]="@test";name[10]="Group:test";
id[11]="&admin";name[11]="Alias:admin";
id[12]="&pm";name[12]="Alias:pm";
id[13]="&tm";name[13]="Alias:tm";
members_count=14;
revision="0.2.1";
""" == res.body, res.body

        params = {'role':'@admin',}
        res = self.app.get(url(controller='role', action='get_role_info'), params)
        assert res.status == "200 OK", res.status
        assert """id[0]="&admin";name[0]="Alias:admin";
id[1]="admin1";name[1]="admin1";
id[2]="admin2";name[2]="admin2";
id[3]="admin3";name[3]="admin3";
members_count=4;
revision="0.2.1";
""" == res.body, res.body

        params = {'role':'@group1',}
        res = self.app.get(url(controller='role', action='get_role_info'), params)
        assert res.status == "200 OK", res.status
        assert """id[0]="@group2";name[0]="Group:group2";
id[1]="@group3";name[1]="Group:group3";
id[2]="user1";name[2]="user1";
id[3]="user11";name[3]="user11";
id[4]="user12";name[4]="user12";
members_count=5;
revision="0.2.1";
""" == res.body, res.body

        params = {'role':'@group3',}
        res = self.app.get(url(controller='role', action='get_role_info'), params)
        assert res.status == "200 OK", res.status
        assert """id[0]="user3";name[0]="user3";
id[1]="user31";name[1]="user31";
id[2]="user32";name[2]="user32";
members_count=3;
revision="0.2.1";
""" == res.body, res.body

        params = {'role':'&admin',}
        res = self.app.get(url(controller='role', action='get_role_info'), params)
        assert res.status == "200 OK", res.status
        assert """aliasname = "&admin";username = "jiangxin";
revision="0.2.1";
""" == res.body, res.body

    def test_save_group(self):
        # authn test
        res = self.app.get(url(controller='role', action='save_group'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/login'), res.location

        # authz test
        self.login('nobody')
        res = self.app.get(url(controller='role', action='save_group'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/security/failed'), res.location


        # Change group members, autodrop=no
        try:
            authz = self.load_authz()
            userobj = authz.get_userobj('@group3')
            self.assert_(unicode(userobj) == u'group3 = user3, user31, user32', unicode(userobj).encode('utf-8'))

            self.login('root')
            params = {'rolename':'group3', 'members':'蒋鑫, user3,@group1', 'autodrop':'no', }
            res = self.app.get(url(controller='role', action='save_group'), params)
            assert res.status == "200 OK", res.status
            assert "Recursive group membership for @group1" in res.body, res.body
        
            authz = self.load_authz()
            userobj = authz.get_userobj('@group3')
            self.assert_(unicode(userobj) == u'group3 = user3, user31, user32', unicode(userobj).encode('utf-8'))
        finally:
            self.rollback()

        # Change group members, autodrop=yes
        try:
            authz = self.load_authz()
            userobj = authz.get_userobj('@group3')
            self.assert_(unicode(userobj) == u'group3 = user3, user31, user32', unicode(userobj).encode('utf-8'))

            self.login('root')
            params = {'rolename':'group3', 'members':'蒋鑫, user3,@group1', 'autodrop':'yes', }
            res = self.app.get(url(controller='role', action='save_group'), params)
            assert res.status == "200 OK", res.status
            assert "" == res.body, res.body
        
            authz = self.load_authz()
            userobj = authz.get_userobj('@group3')
            self.assert_(unicode(userobj) == u'group3 = user3, 蒋鑫', unicode(userobj).encode('utf-8'))
        finally:
            self.rollback()

        # Add New Group
        try:
            authz = self.load_authz()
            userobj = authz.get_userobj('@管理员组')
            self.assert_(userobj == None)

            self.login('root')
            params = {'rolename':'管理员组', 'members':'蒋鑫, user3,@group1' }
            res = self.app.get(url(controller='role', action='save_group'), params)
            assert res.status == "200 OK", res.status
            assert "" == res.body, res.body
        
            authz = self.load_authz()
            userobj = authz.get_userobj('@管理员组')
            self.assert_(unicode(userobj) == u'管理员组 = @group1, user3, 蒋鑫', unicode(userobj).encode('utf-8'))
        finally:
            self.rollback()

        # Add New Group, out-of-date
        try:
            authz = self.load_authz()
            userobj = authz.get_userobj('@管理员组')
            self.assert_(userobj == None)

            self.login('root')
            params = {'rolename':'管理员组', 'members':'蒋鑫, user3,@group1', 'revision':'' }
            res = self.app.get(url(controller='role', action='save_group'), params)
            assert res.status == "200 OK", res.status
            assert "Update failed! You are working on a out-of-date revision." in res.body, res.body
        
            authz = self.load_authz()
            userobj = authz.get_userobj('@管理员组')
            self.assert_(userobj == None)
        finally:
            self.rollback()
            
    def test_delete_group(self):
        # authn test
        res = self.app.get(url(controller='role', action='delete_group'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/login'), res.location

        # authz test
        self.login('nobody')
        res = self.app.get(url(controller='role', action='delete_group'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/security/failed'), res.location

        # Delete group failed, ref by other group.
        try:
            authz = self.load_authz()
            userobj = authz.get_userobj('@group3')
            self.assert_(unicode(userobj) == u'group3 = user3, user31, user32', unicode(userobj).encode('utf-8'))

            self.login('root')
            params = {'role':'group3',}
            res = self.app.get(url(controller='role', action='delete_group'), params)
            assert res.status == "200 OK", res.status
            assert "Group group3 is referenced by group @group1." in res.body, res.body
        finally:
            self.rollback()

        # Delete group failed, ref by rules.
        try:
            authz = self.load_authz()
            userobj = authz.get_userobj('@dev')
            self.assert_(unicode(userobj) == u'dev = dev1, dev2, dev3', unicode(userobj).encode('utf-8'))

            self.login('root')
            params = {'role':'@dev',}
            res = self.app.get(url(controller='role', action='delete_group'), params)
            assert res.status == "200 OK", res.status
            assert "@dev is referenced by [/:/trunk]." in res.body, res.body
        
            authz = self.load_authz()
            userobj = authz.get_userobj('@dev')
            self.assert_(unicode(userobj) == u'dev = dev1, dev2, dev3', unicode(userobj).encode('utf-8'))
        finally:
            self.rollback()

        # Delete group successful.
        try:
            authz = self.load_authz()
            userobj = authz.get_userobj('@all')
            self.assert_(unicode(userobj) == 'all = @admin, @dev, @test', unicode(userobj).encode('utf-8'))

            self.login('root')
            params = {'role':'all',}
            res = self.app.get(url(controller='role', action='delete_group'), params)
            assert res.status == "200 OK", res.status
            assert "" == res.body, res.body
        
            authz = self.load_authz()
            userobj = authz.get_userobj('@all')
            self.assert_(userobj==None)
        finally:
            self.rollback()

    def test_save_alias(self):
        # authn test
        res = self.app.get(url(controller='role', action='save_alias'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/login'), res.location

        # authz test
        self.login('nobody')
        res = self.app.get(url(controller='role', action='save_alias'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/security/failed'), res.location

        # Change alias successfully
        try:
            authz = self.load_authz()
            userobj = authz.get_userobj('&admin')
            self.assert_(unicode(userobj) == u'admin = jiangxin', unicode(userobj).encode('utf-8'))

            self.login('root')
            params = {'aliasname':'admin', 'username':'蒋鑫',}
            res = self.app.get(url(controller='role', action='save_alias'), params)
            assert res.status == "200 OK", res.status
            assert "" == res.body, res.body
        
            authz = self.load_authz()
            userobj = authz.get_userobj('&admin')
            self.assert_(unicode(userobj) == u'admin = 蒋鑫', unicode(userobj).encode('utf-8'))
            self.assert_(authz.is_super_user('&admin')==True, authz.is_super_user('&admin'))
            self.assert_(authz.is_super_user('蒋鑫')==True, authz.is_super_user('蒋鑫'))
            
            self.login('蒋鑫')
            #params = {'aliasname':'admin', 'username':'蒋鑫',}
            params = {'aliasname':'admin2', 'username':'jiangxin',}
            res = self.app.get(url(controller='role', action='save_alias'), params)
            assert res.status == "200 OK", res.status
            assert "" == res.body, res.body
            
        finally:
            self.rollback()
            
        # Add new alias successfully
        try:
            authz = self.load_authz()
            userobj = authz.get_userobj('&管理员')
            self.assert_(userobj == None)

            self.login('root')
            params = {'aliasname':'管理员', 'username':'蒋鑫',}
            res = self.app.get(url(controller='role', action='save_alias'), params)
            assert res.status == "200 OK", res.status
            assert "" == res.body, res.body
        
            authz = self.load_authz()
            userobj = authz.get_userobj('&管理员')
            self.assert_(unicode(userobj) == u'管理员 = 蒋鑫', unicode(userobj).encode('utf-8'))
        finally:
            self.rollback()
            
        # Change alias failed, out-of-date
        try:
            authz = self.load_authz()
            userobj = authz.get_userobj('&admin')
            self.assert_(unicode(userobj) == u'admin = jiangxin', unicode(userobj).encode('utf-8'))

            self.login('root')
            params = {'aliasname':'admin', 'username':'蒋鑫', 'revision':'123'}
            res = self.app.get(url(controller='role', action='save_alias'), params)
            assert res.status == "200 OK", res.status
            assert "Update failed! You are working on a out-of-date revision." in res.body, res.body
        
            authz = self.load_authz()
            userobj = authz.get_userobj('&admin')
            self.assert_(unicode(userobj) == u'admin = jiangxin', unicode(userobj).encode('utf-8'))
        finally:
            self.rollback()
            
            
    def test_delete_alias(self):
        # authn test
        res = self.app.get(url(controller='role', action='delete_alias'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/login'), res.location

        # authz test
        self.login('nobody')
        res = self.app.get(url(controller='role', action='delete_alias'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/security/failed'), res.location

        # Delete alias successfully
        try:
            authz = self.load_authz()
            userobj = authz.get_userobj('&tm')
            self.assert_(unicode(userobj) == u'tm = test1', unicode(userobj).encode('utf-8'))

            self.login('root')
            params = {'aliasname':'tm'}
            res = self.app.get(url(controller='role', action='delete_alias'), params)
            assert res.status == "200 OK", res.status
            assert "" == res.body, res.body
        
            authz = self.load_authz()
            userobj = authz.get_userobj('&tm')
            self.assert_(userobj == None)
        finally:
            self.rollback()
            
        # Delete alias failed. out of date
        try:
            authz = self.load_authz()
            userobj = authz.get_userobj('&tm')
            self.assert_(unicode(userobj) == u'tm = test1', unicode(userobj).encode('utf-8'))

            self.login('root')
            params = {'aliasname':'tm', 'revision':''}
            res = self.app.get(url(controller='role', action='delete_alias'), params)
            assert res.status == "200 OK", res.status
            assert "Update failed! You are working on a out-of-date revision." in res.body, res.body
        
            authz = self.load_authz()
            userobj = authz.get_userobj('&tm')
            self.assert_(unicode(userobj) == u'tm = test1', unicode(userobj).encode('utf-8'))
        finally:
            self.rollback()
            

        # Delete alias failed. refenced by rules.
        try:
            authz = self.load_authz()
            userobj = authz.get_userobj('&pm')
            self.assert_(unicode(userobj) == u'pm = dev1', unicode(userobj).encode('utf-8'))

            self.login('root')
            params = {'aliasname':'pm', 'revision':''}
            res = self.app.get(url(controller='role', action='delete_alias'), params)
            assert res.status == "200 OK", res.status
            assert "&pm is referenced by [/:/trunk]." in res.body, res.body
        
            authz = self.load_authz()
            userobj = authz.get_userobj('&pm')
            self.assert_(unicode(userobj) == u'pm = dev1', unicode(userobj).encode('utf-8'))
        finally:
            self.rollback()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2008 OpenSourceXpress Ltd. (http://www.ossxp.com)
# Author: Jiang Xin
# Contact: http://www.ossxp.com
#          http://www.worldhello.net
#          http://moinmo.in/JiangXin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

"""Subversion authz config file management.

Basic classes used for Subversion authz management.
"""

from configobj import ConfigObj
import rcsbackup as rcs
import re
import sys
import os
import my_fnmatch
import StringIO
import logging
log = logging.getLogger(__name__)

# i18n works only as pysvnmanager (a pylons app) model.
from pylons import config
if config.has_key('unittest'):
    from pylons.i18n import _
else:
    def _(message): return message

#reload(sys) # in Python2.5, method sys.setdefaultencoding 
            #will be delete after initialize. we need reload it.
#sys.setdefaultencoding('utf-8')

RIGHTS_W = 4
RIGHTS_R = 2
RIGHTS_NONE = 0
RIGHTS_RW = RIGHTS_R | RIGHTS_W
RIGHTS_ALL = RIGHTS_R | RIGHTS_W

def check_valid_username(name):
    bad_chars = r'''^[\.$&@~-]|[\\/:]|[*?"'<>|,;%#$]'''
    return check_valid_string(name, bad_chars=bad_chars)

def check_valid_aliasname(name):
    return check_valid_username(name)

def check_valid_groupname(name):
    if name == '*':
        return
    else:
        if name[0]=='$':
            name = name[1:]
        return check_valid_username(name)

def check_valid_reposname(name):
    if name == '/':
        return
    else:
        bad_chars = r'''^[\.$&@-]|[\\/:]|[*?"'<>|,;%#$]'''
        return check_valid_string(name, bad_chars=bad_chars)
    
def check_valid_string(check_str, bad_chars=""):
    check_str = normalize_user(check_str)
    msg = ''
    if not check_str:
        msg = _("Name is not given.")
    elif not isinstance(check_str, basestring):
        msg = _("Name is not string.")
    else:
        p = re.compile(bad_chars)
        if p.search(check_str):
            msg = _(u"Name (%s) contains invalid characters.") % check_str

    if msg:
        raise Exception, msg

def normalize_user(name):
    if isinstance(name, basestring):
        name = name.strip()

    if isinstance(name, str):
        name = unicode(name, 'utf-8')

    return name

def normalize_repos(name):
    if isinstance(name, basestring):
        name = name.strip()
        name = name.strip('/')
    if not name:
        name = '/'

    if isinstance(name, str):
        name = unicode(name, 'utf-8')

    return name

def normalize_path(name):
    if isinstance(name, basestring):
        name = name.strip()
        name = name.strip('/')
    if not name:
        name = '/'
    elif isinstance(name, basestring) and name[0] != '/':
        name = '/'+name
    
    if isinstance(name, str):
        name = unicode(name, 'utf-8')

    return name

class User(object):
    """User object used in groups, alias, and rules, etc.
    
    Useful attributes:
        name  : user login name.
        uname : unique name. compatible with Group, Alias objects.

    Method:
        attributes:

    >>> user = User('Jiang Xin')
    >>> user.name
    'Jiang Xin'
    >>> user.uname
    'Jiang Xin'
    >>> User('Jiang Xin') in user
    True
    >>> user in user
    True
    >>> 'Jiang Xin' in user
    True
    >>> 'jiang xin' in user
    True
    >>> user = User('')
    Traceback (most recent call last):
            ...
    Exception: Username is not provided
    """

    def __init__(self, name):
        name = normalize_user(name)

        if not name:
            raise Exception, 'Username is not provided'

        self.__name = name

    def __get_unique_name(self):
        return self.__name

    uname = property(__get_unique_name)
    name  = property(__get_unique_name)

    def __cmp__(self, obj):
        """For userlist sorting"""
        if isinstance(obj, User):
            return cmp(self.uname, obj.uname)
        else:
            return -1

    def __contains__(self, obj):
        """Match if is the same user.

        Compatible with Alias and Groups's __contains__ method."""
        if isinstance(obj, User):
            obj = obj.uname
        elif isinstance(obj, Alias):
            obj = obj.username
        elif isinstance(obj, (str, unicode)):
            obj = normalize_user(obj)
        else:
            return False

        if not obj:
            return False

        return obj.lower() == self.uname.lower()
    
    def __str__(self):
        return self.__name


class Alias(object):
    """Alias object used in groups, and rules.
    
    Alias is defined in svnauthz file like this:
        [aliases]
        aliasname = realuser
    
    Alias can be referenced in Group and rules as:
        [group]
        team1 = user1, &aliasname, @team2

        [repos:/trunk]
        * = 
        &aliasname = rw
    
    >>> user = User('username')
    >>> a = Alias('aliasname', user)
    >>> a.uname
    '&aliasname'
    >>> unicode(a)
    'aliasname = username'
    >>> user in a
    True
    >>> user = User('jiangxin')
    >>> a.user = user
    >>> unicode(a)
    'aliasname = jiangxin'
    >>> 'JiangXin' in a
    True
    >>> 'aliasname' in a
    False
    >>> '&aliasname' in a
    True
    >>> a in a
    True
    >>> a = Alias('admin')
    >>> unicode(a)
    'admin = '
    >>> a.username
    ''
    >>> '&admin' in a
    True
    >>> Alias('&aliasname')
    Traceback (most recent call last):
            ...
    Exception: Aliasname should not begin with &.
    >>> Alias('')
    Traceback (most recent call last):
            ...
    Exception: Aliasname is not provided.
    """

    def __init__(self, aliasname, userobj=None):
        if userobj != None and (not isinstance(userobj, User)) :
            raise Exception, 'Wrong parameter for userobj.'

        aliasname = normalize_user(aliasname)
        check_valid_aliasname(aliasname)
        self.__name = aliasname
        self.__userobj = userobj


    def __get_name(self):
        return self.__name

    name = property(__get_name)
    aliasname = property(__get_name)

    def __get_unique_name(self):
        return u'&'+self.__name

    uname = property(__get_unique_name)

    def __get_username(self):
        if self.__userobj:
            return self.__userobj.name
        else:
            return u''

    username = property(__get_username)

    def __set_user(self, userobj):
        if not isinstance(userobj, User):
            raise Exception, 'Wrong parameter for userobj.'
        self.__userobj = userobj

    user = property(__get_username, __set_user)

    def __str__(self):
        return u"%s = %s" % (self.aliasname, self.username)

    def __cmp__(self, obj):
        """For list sorting"""
        if isinstance(obj, Alias):
            return cmp(self.uname, obj.uname)
        else:
            return -1

    def __contains__(self, obj):
        if isinstance(obj, Alias):
            obj = obj.username
        elif isinstance(obj, (User, Group)):
            obj = obj.uname
        elif isinstance(obj, basestring):
            obj = normalize_user(obj)

        if not obj:
            return False

        obj = obj.lower()
        if obj == self.uname.lower() or \
              obj == self.username.lower():
            return True

        return False


class Group(object):
    """Group objects. Referenced in groups, and rules.
    
    Group is defined in svnauthz file like this:
        [group]
        team1 = user1, user2, &aliasname
    
    Group can be referenced in Group and rules as:
        [group]
        team1 = user1, user2, &aliasname, @team2
        team2 = user3, user4 

        [repos:/trunk]
        * = 
        @team1 = rw

    >>> authz = SvnAuthz()
    >>> user1 = authz.add_user('user1')
    >>> user2 = authz.add_user('user2')
    >>> user3 = authz.add_user('user3')
    >>> user4 = authz.add_user('user4')
    >>> jiangxin = authz.add_user('jiangxin')
    >>>
    >>> admin = authz.add_alias('admin', jiangxin)
    >>> N007  = authz.add_alias('N007', user4)
    >>>
    >>> admins = authz.add_group('admins')
    >>> team1  = authz.add_group('team1')
    >>> team2  = authz.add_group('team2')
    >>> team3  = authz.add_group('team3')
    >>> all    = authz.add_group('all')
    >>>
    >>> authz.add_group_member(admins, '&admin, &N007')
    ... # doctest: +ELLIPSIS
    <...object at ...>
    >>> authz.add_group_member(team1, [user1, team2]) 
    ... # doctest: +ELLIPSIS
    <...Group object at ...>
    >>> authz.add_group_member(team2, [user2, team3])
    ... # doctest: +ELLIPSIS
    <...Group object at ...>
    >>> authz.add_group_member(team3, [user3, team1], autodrop=True)
    ... # doctest: +ELLIPSIS
    <...Group object at ...>
    >>> authz.add_group_member(all, [team1, user3, user4])
    ... # doctest: +ELLIPSIS
    <...Group object at ...>
    >>>
    >>> print authz # doctest: +NORMALIZE_WHITESPACE
    # version : 0.0
    <BLANKLINE>
    [groups]
    admins = &N007, &admin
    all = @team1, user3, user4
    team1 = @team2, user1
    team2 = @team3, user2
    team3 = user3
    <BLANKLINE>
    [aliases]
    N007 = user4
    admin = jiangxin
    <BLANKLINE>
    <BLANKLINE>
    >>> jiangxin in admins
    True
    >>> 'user4' in admins
    True
    >>> 'user1' in admins
    False
    >>> '@team1' in team3
    False
    >>> 'user3' in team1
    True
    >>> star = authz.add_group('*')
    >>> anon = authz.add_group('$anonymous')
    >>> auth = authz.add_group('$authenticated')
    >>> print star
    # Built-in group: *
    >>> print anon
    # Built-in group: $anonymous
    >>> print auth
    # Built-in group: $authenticated
    >>> '*' in star
    True
    >>> 'valid_user' in star
    True
    >>> '$anonymous' in star
    True
    >>> '$authenticated' in star
    True
    >>> '*' in anon
    True
    >>> 'valid_user' in anon
    False
    >>> '$anonymous' in anon
    True
    >>> '$authenticated' in anon
    False
    >>> '*' in auth
    False
    >>> 'valid_user' in auth
    True
    >>> '$anonymous' in auth
    False
    >>> '$authenticated' in auth
    True
    """

    def __init__(self, name=''):
        name = normalize_user(name)
        check_valid_groupname(name)
        self.name = name
        self.__members = []

    def __get_unique_name(self):
        if self.name[0] != '$' and self.name != '*':
            return u'@'+self.name
        else:
            return self.name

    uname = property(__get_unique_name)

    def __get_member_names(self):
        if self.__members:
            return sorted(map(lambda x: x.uname, self.__members))
        else:
            return []

    membernames = property(__get_member_names)

    def __get_member_objs(self):
        return self.__members

    memberobjs = property(__get_member_objs)

    def __str__(self):
        if self.name[0] != '$' and self.name != '*':
            return u"%s = %s" % (self.name, ', '.join(sorted(self.membernames)))
        else:
            return u"# Built-in group: %s" % self.name

    def __iter__(self):
        for i in self.__members:
            yield i
    
    def __valid_group(self, groupobj, checked_groups=None):
        """x.valid_group(groupobj) -> bool

        Check if there is cycle reference in group defination."""
        if not checked_groups:
            checked_groups = [groupobj]
        for member in groupobj.memberobjs:
            if not isinstance(member, Group):
                continue
            if member in checked_groups:
                return False
            # Add group to hash of checked groups.
            checked_groups.append(member)
            # Check for deadly loop recursively.
            if not self.__valid_group(member, checked_groups):
                return False
            # Remove group to hash of checked groups. 
            # Other group members may contain this group
            checked_groups.remove(member)
        return True

    def append(self, *members, **opts):
        '''**opts: autodrop=False, raise Exception if recursive found.'''
        autodrop = opts.get('autodrop', False)
        for member in members:
            if isinstance(member, (list, tuple)):
                for i in member:
                    self.append(i, autodrop=autodrop)
                continue
            if not member in self.__members:
                self.__members.append(member)
                if not self.__valid_group(self):
                    self.__members.remove(member)
                    if not autodrop:
                        raise Exception, _('Recursive group membership for %s') \
                            % member.uname

    def remove(self, *members):
        ulist = []
        isRemoved = False
        for member in members:
            if isinstance(member, (str, unicode)):
                for user in member.split(','):
                    user = normalize_user(user)
                    if user:
                        ulist.append(user)
            elif isinstance(member, (list, tuple)):
                ulist.extend(member)
        
        mlist = map(lambda x: x.uname, self.__members)
        for user in ulist:
            if isinstance(user, (str, unicode)):
                user = normalize_user(user)
                if user not in mlist:
                    continue
                del self.__members[mlist.index(user)]
                isRemoved = True
            else:
                if user in self.__members:
                    self.__members.remove(user)
                    isRemoved = True
            mlist = map(lambda x: x.uname, self.__members)
        
        return isRemoved

    def remove_all(self):
        self.__members = []

    def __cmp__(self, obj):
        """For list sorting"""
        if isinstance(obj, Group):
            return cmp(self.uname, obj.uname)
        else:
            return -1

    def __contains__(self, obj):

        if isinstance(obj, Alias):
            obj = obj.username
        elif isinstance(obj, (User, Group)):
            obj = obj.uname
        elif isinstance(obj, (str, unicode)):
            obj = normalize_user(obj)
        else:
            return False

        if not obj:
            return False

        obj = obj.lower()
        if obj == self.uname.lower():
            return True

        if self.uname == '*':
            return True
        if self.uname == '$anonymous' and obj == '*':
            return True
        if self.uname == '$authenticated':
            if obj != '$anonymous' and obj != '*':
                return True

        for member in self.__members:
            if obj in member:
                return True

        return False


class UserList(object):
    """Store all users referenced by [group], [aliases], [repos:/a/b] sections.
    """
    def __init__(self):
        self._user_list = []

    def __get_user_list(self):
        self._user_list.sort()
        return self._user_list

    user_list = property(__get_user_list)

    def __iter__(self):
        for i in self.user_list:
            yield i

    def get(self, name):
        return self.get_or_set(name, autocreate=False)

    def get_or_set(self, name, autocreate = True):
        assert isinstance(name, basestring)
        
        name = normalize_user(name)
        if not name:
            return None
        if name[0] == '&' or name[0] == '@' or name[0] == '$' or name=='*':
            raise Exception, _("Not a valid username: %s") % name

        for user in self._user_list:
            if user.name == name:
                return user

        if autocreate:
            check_valid_username(name)
            user = User(name)
            self._user_list.append(user)
            return user
        else:
            return None


class AliasList(object):
    """Store all alias objects defined by [aliases] section or referenced in 
    [group], [repos:/path/to] but not be set.
    """
    def __init__(self):
        self._alias_list = []

    def __get_alias_list(self):
        self._alias_list.sort()
        return self._alias_list

    alias_list = property(__get_alias_list)

    def __iter__(self):
        for i in self.alias_list:
            yield i

    def get(self, name):
        return self.get_or_set(name, False)

    def get_or_set(self, name, autocreate = True):
        assert isinstance(name, basestring)
        name = normalize_user(name)

        if not name:
            return None
        if name[0] == '&':
            name = name[1:]
        for alias in self._alias_list:
            if alias.aliasname == name:
                return alias

        if autocreate:
            check_valid_aliasname(name)
            alias = Alias(name)
            self._alias_list.append(alias)
            return alias
        else:
            return None

    def remove(self, name):
        if isinstance(name, (str, unicode)):
            alias = self.get(name)
        else:
            alias = name
        if alias and alias in self._alias_list:
            self._alias_list.remove(alias)
            return True
        else:
            return False

    def __str__(self):
        buff = u"[aliases]\n"
        for alias in self.alias_list:
            buff += unicode(alias)
            buff += u'\n'
        return buff


class GroupList(object):
    """Store all group objects defined by [groups] section or referenced in 
    [groups], [repos:/path/to] but not be set.
    """
    def __init__(self):
        self._group_list = []

    def __iter__(self):
        for i in self.group_list:
            yield i

    def __get_group_list(self):
        self._group_list.sort()
        return self._group_list

    group_list = property(__get_group_list)

    def get(self, name):
        return self.get_or_set(name, False)

    def get_or_set(self, name, autocreate = True):
        assert isinstance(name, basestring)
        name = normalize_user(name)
        if not name:
            return None
        if name[0] == '@':
            name = name[1:]
        for group in self._group_list:
            if group.name == name:
                return group

        if autocreate:
            check_valid_groupname(name)
            group = Group(name)
            self._group_list.append(group)
            return group
        else:
            return None

    def remove(self, name, force=False):
        if isinstance(name, Group):
            name = name.name
        else:
            name = normalize_user(name)
        item = self.get(name)
        if item:
            for group in self._group_list:
                if group == item:
                    continue
                if item in group.memberobjs:
                    if force:
                        group.memberobjs.remove(item)
                        assert not item in group.memberobjs
                    else:
                        raise Exception, \
                                _('Group %s is referenced by group %s.') \
                                % (name, group.uname)
            self._group_list.remove(item)
            return True
        else:
            return False

    def __str__(self):
        buff = u"[groups]\n"
        for group in self.group_list:
            if group.name[0] != '$' and group.name != '*':
                buff += unicode(group)
                buff += u'\n'
        return buff


class Rule(object):
    """Store one single svnauthz rule.

    Attribute:
        userobj -- points to a user/group/alias object. Has a match method, 
                              which can detect whether the rule hits certain 
                              user/group/alias object.
        uname   -- the same as userobj's uname attribute.
        rights  -- for convenience allow two different ways to save rights:
                                  Set the rights using str: as 'r', 'rw', ''. or 
                                  Set with raw bits value: RIGHTS_R, RIGHTS_W, ...
                              Rights property return only raw bits, remember.
    """
    def __init__(self, userobj):
        self.userobj = userobj
        self.__rights = RIGHTS_NONE

    def __get_unique_name(self):
        return self.userobj.uname

    uname = property(__get_unique_name)

    def __set_rights(self, rights):
        if isinstance(rights, (str, unicode)):
            rights = rights.strip().lower()
            rbit = RIGHTS_NONE
            if 'r' in rights:
                rbit |= RIGHTS_R
            if 'w' in rights:
                rbit |= RIGHTS_W
            rights = rbit
        else:
            rights = int(rights) & RIGHTS_ALL

        self.__rights = rights

    def __get_rights(self):
        return self.__rights

    rights = property(__get_rights, __set_rights)

    def get_permission(self, userobj):
        """x.get_permission(userobj) -> (permit_bits, deny_bits)

        If userobj match with rule's userobj, return permit and deny bits.
        Else return (ZERO, ZERO)
        """
        if not userobj in self.userobj:
            return (RIGHTS_NONE, RIGHTS_NONE)

        perm = self.rights
        deny = RIGHTS_ALL ^ perm
        return (perm, deny)

    def __str__(self):
        rstr = u''
        rbit = self.__rights
        if rbit & RIGHTS_R:
            rstr += u'r'
        if rbit & RIGHTS_W:
            rstr += u'w'
        return u"%s = %s" % (self.userobj.uname, rstr)

    def __cmp__(self, obj):
        """For list sorting"""
        if isinstance(obj, Rule):
            return cmp(self.userobj.uname, obj.userobj.uname)
        else:
            return -1


class Module(object):
    """Module stores one svnauthz module config section as [repos:/path/to].

    A module has a repos name store in repos, and path of the module.
    Also the main part is rule_list[], store a list of Rule objects
    for this module.
    """
    def __init__(self, repos, path):
        self.repos = normalize_repos(repos)
        self.path  = normalize_path(path)
        self.__rule_list  = []

    def __iter__(self):
        for i in self.__rule_list:
            yield i
    
    def __get_fullame(self):
        return self.repos+':'+self.path

    fullname = property(__get_fullame)

    def clean_rules(self):
        self.__rule_list = []

    def del_rule(self, rules):
        if isinstance(rules, (str, unicode)):
            rdict = {}
            key, value = rules.split('=', 1)
            rdict[key.strip()] = value.strip()
            return self.del_rule(rdict)
        elif isinstance(rules, (list, tuple)):
            for rule in rules:
                self.del_rule(rule)
            return True

        if not isinstance(rules, dict):
            raise

        unamelist = map(lambda x: x.uname, self.__rule_list)
        for (user, rights) in rules.items():
            user = normalize_user(user)

            if user in unamelist:
                del self.__rule_list[unamelist.index(user)]
                unamelist = map(lambda x: x.uname, self.__rule_list)

        return True

    def update_rule(self, obj, rights):
        unamelist = map(lambda x: x.uname, self.__rule_list)
        if obj.uname in unamelist:
            rule = self.__rule_list[unamelist.index(obj.uname)]
        else:
            rule = Rule(obj)
            self.__rule_list.append(rule)
        rule.rights = rights

        return True

    def __get_rules(self):
        return self.__rule_list
   
    rules = property(__get_rules)
    
    def __str__(self):
        if not self.__rule_list:
            return u''
        if self.repos == '/' or not self.repos:
            buff = u"[%s]\n" % self.path
        else:
            buff = u"[%s:%s]\n" % (self.repos, self.path)
        for rule in sorted(self.__rule_list):
            tmp = unicode(rule)
            if tmp:
                buff += tmp
                buff += u'\n'
        return buff

    def __cmp__(self, obj):
        """For list sorting"""
        if isinstance(obj, Module):
            return cmp(self.fullname, obj.fullname)
        else:
            return -1

    def get_permit_bits(self, user):
        """x.get_permit_bits(user) -> (perm, deny)
        
        Return permission bits of this module as a tuple.

        Warning: the return value is meaningful only if access_is_determined 
        returns True. Check return value of access_is_determined first!
        """
        perm = RIGHTS_NONE
        deny = RIGHTS_NONE
        for rule in self.__rule_list:
            pbits, dbits = rule.get_permission(user)
            perm |= pbits
            deny |= dbits
        return (perm, deny)

    def get_permit_str(self, user):
        """x.get_permit_str(user) -> string
        
        Return user friendly permission of this module.

        Warning: the return value is meaningful only if access_is_determined 
        returns True. Check return value of access_is_determined first!
        """
        permstr = ''
        perm, deny = self.get_permit_bits(user)
        if perm & RIGHTS_R:
            permstr += 'r'
        if perm & RIGHTS_W:
            permstr += 'w'
        return permstr

    def access_is_determined(self, user):
        """x.access_is_determined(user) -> bool
        
        Whether a certain user object matches with some of 
        the rules for this module.

        Permission return by other funtions is meaningful only if the method
        returns True.
        """
        perm, deny = self.get_permit_bits(user)
        if (RIGHTS_ALL & perm) or (RIGHTS_ALL & deny):
            return True
        else:
            return False

    def access_is_granted(self, user, required):
        """x.access_is_granted(user, required) -> bool
        
        Whether the user required rights is granted by rules of this module.

        Warning: the return value is meaningful only if access_is_determined 
        returns True. Check return value of access_is_determined first!
        """
        if isinstance(required, (str, unicode)):
            required = required.strip().lower()
            rbit = RIGHTS_NONE
            if 'r' in required:
                rbit |= RIGHTS_R
            if 'w' in required:
                rbit |= RIGHTS_W
            required = rbit
        else:
            required = int(required) & RIGHTS_ALL
        perm, deny = self.get_permit_bits(user)
        if deny & required == RIGHTS_NONE:
            return True
        elif perm & required == required & RIGHTS_ALL:
            return True
        else:
            return False


class Repos(object):
    """One repos may contain many modules.

    These modules share the same repos name, and have the same administrator(s).
    """
    def __init__(self, name):
        name = normalize_repos(name)
        self.__repos_name = name
        self.__admins = []
        self.module_list = []

    def __iter__(self):
        for i in self.module_list:
            yield i
    
    def __get_name(self):
        return self.__repos_name

    def __set_name(self, name):
        self.__repos_name = normalize_repos(name)

    name = property(__get_name, __set_name)

    def __get_path_list(self):
        return map(lambda x:x.path, self.module_list)
    
    path_list = property(__get_path_list)

    def __get_admins(self):
        alist = [i.uname for i in self.__admins]
        return ', '.join(sorted(alist))

    def __set_admins(self, admins):
        self.__admins = []
        return self.add_admin(admins)

    admins = property(__get_admins, __set_admins)
    
    def add_admin(self, admin):
        """x.add_admin(admin)"""
        if isinstance(admin, (User, Group, Alias)):
            if not admin in self.__admins:
                self.__admins.append(admin)
        elif isinstance(admin, (list, tuple, set)):
            for i in admin:
                self.add_admin(i)
        else:
            raise Exception, "unknown user: %s, type: %s" % (admin, type(admin))

    def del_admin(self, admin):
        if isinstance(admin, (list, tuple, set)):
            for i in admin:
                self.del_admin(i)
        elif isinstance(admin, (User, Group, Alias)):
            self.__admins.remove(admin)
        else:
            raise Exception, "unknown user: %s, type: %s" % (admin, type(admin))
    
    def add_module(self, path):
        path = normalize_path(path)

        for i in self.module_list:
            if i.path.lower() == path.lower():
                return i
        i = Module(repos=self.name , path=path)
        self.module_list.append(i)
        return i

    def del_module(self, path):
        path = normalize_path(path)

        mlist = map(lambda x: x.path, self.module_list)
        if path in mlist:
            del self.module_list[mlist.index(path)]
            return True
        else:
            return False

    def del_all_modules(self):
        self.module_list = []

    def get_module(self, path):
        path = normalize_path(path)

        for i in self.module_list:
            if i.path.lower() == path.lower():
                return i
        # OSSXP hacked subversion supports wildcard characters as module path.
        for i in self.module_list:
            if '*' in i.path or '?' in i.path:
                if my_fnmatch.fnmatch(path, i.path):
                    return i
        return None

    def is_blank(self):
        if self.__admins or self.module_list:
            return False
        else:
            return True

    def __str__(self):
        buff = u''
        for i in sorted(self.module_list):
            tmp = unicode(i)
            if tmp:
                buff += tmp
                buff += u'\n'
        return buff

    def __cmp__(self, obj):
        """For list sorting"""
        if isinstance(obj, Repos):
            return cmp(self.name, obj.name)
        else:
            return -1


class ReposList(object):
    def __init__(self):
        self.repos_list = []
        self.get_or_set('/')

    def __iter__(self):
        for i in self.repos_list:
            yield i

    def get(self, name):
        return self.get_or_set(name, False)

    def get_or_set(self, name, autocreate = True):
        name = normalize_repos(name)
        assert isinstance(name, basestring)

        for repos in self.repos_list:
            if repos.name.lower() == name.lower():
                return repos

        if autocreate:
            check_valid_reposname(name)
            repos = Repos(name)
            self.repos_list.append(repos)
            return repos
        else:
            return None

    def remove(self, name, recursive=False):
        if isinstance(name, Repos):
            name = name.name

        name = normalize_repos(name)

        if name == '/' and not recursive:
            return False

        rlist = map(lambda x: x.name, self.repos_list)
        try:
            idx = rlist.index(name)
        except:
            raise Exception, "Repos '%s' not exist" % name

        repos = self.repos_list[idx]
        assert(repos.name.lower() == name.lower())

        if recursive:
            if name == '/':
                self.repos_list[idx].del_all_modules()
            else:
                del self.repos_list[idx]
            return True

        else:
            if repos.is_blank():
                del self.repos_list[idx]
                return True
            else:
                return False

    def __str__(self):
        buff = u''
        for repos in sorted(self.repos_list):
            tmp = unicode(repos)
            if tmp:
                buff += tmp
                ##buff += '\n'
        return buff


class SvnAuthz(object):
    '''
    Data structure:
    * SvnAuthz
            * UserList  user_list
                    * User  user_list[]
            * AliasList alias_list
                    * Alias alias_list[]
            * GroupList group_list
                    * Group group_list[]
            * ReposList repos_list
                    * Repos repos_list[]
                            str __repos_name
                            __admins = set(str[])
                            * Module module_list[]
                                    str repos
                                    str path
                                    * Rule rule_list[]
                                          Group/Alias/User id
                                          str right
    '''
    def __init__(self, fileobj=None):
        self.__clear()
        self.__file = None
        # Used as check-in username to rcs file.
        self.login_as = None
        # Used as check-in message to rcs file.
        self.comment = []
        self.load(fileobj)
    
    def __clear(self):
        self.__userlist  = UserList()
        self.__aliaslist = AliasList()
        self.__grouplist = GroupList()
        self.__reposlist = ReposList()
        self.__version = '0.0'
        self.config = None
        self.add_repos('/')        

    def __get_userlist(self):
        return self.__userlist

    userlist = property(__get_userlist)

    def __get_aliaslist(self):
        return self.__aliaslist

    aliaslist = property(__get_aliaslist)

    def __get_grouplist(self):
        return self.__grouplist

    grouplist = property(__get_grouplist)

    def __get_reposlist(self):
        return self.__reposlist

    reposlist = property(__get_reposlist)

    def __get_version(self):
        return self.__version
    
    version = property(__get_version)
    
    def update_revision(self):
        major, minor = self.version.rsplit('.',1)
        if minor.isdigit():
            rev = int(minor)+1
        else:
            rev = 0
        self.__version = u"%s.%d" % (major, rev)
        
    def modulelist(self):
        for i in self.reposlist:
            for j in i.module_list:
                yield j

    def rulelist(self):
        for i in self.reposlist:
            for j in i.module_list:
                for k in j:
                    yield k

    def load(self, fileobj=None):
        '''
        Initial SvnAuthz from authz file.
        file can be filename, or file handler, StingIO,...
        '''
        self.__clear()
        
        self.__file = fileobj
        if self.__file:
            assert isinstance(self.__file, (basestring, file, StringIO.StringIO))
            # set encoding to 'utf8'
            self.config = ConfigObj(self.__file, encoding='utf8', default_encoding='utf8')

            if self.config:
                self.parse_acl()
                self.parse_version()
                if self.config.has_key('groups'):
                    self.parse_groups(self.config['groups'])
                if self.config.has_key('aliases'):
                    self.parse_aliases(self.config['aliases'])
                for (section, contents) in self.config.items():
                    if section == 'groups' or section == 'aliases':
                        continue
                    self.parse_module(section, contents)

    def save(self, revision, comment=""):
        if comment: self.comment.append(comment)
            
        if self.__file:
            assert isinstance(self.__file, (basestring, StringIO.StringIO))
            #if not revision:
            #    revision = self.version
            last_rev = self.get_revision_from_file()
            log.debug("this revision: %s, last: %s" % (revision, last_rev))
            if last_rev and revision != last_rev:
                raise Exception, _("Update failed! You are working on a out-of-date revision.") + " (%s <> %s)" % (revision, last_rev)

            self.update_revision()

            if isinstance(self.__file, basestring):
                f = open(self.__file, 'w')
            else:
                f = self.__file
                f.truncate(0)
            f.write(unicode(self).encode('utf-8'))
            if isinstance(self.__file, (basestring)):
                f.close()
            else:
                f.seek(0,0)
                f.flush()

            if isinstance(self.__file, basestring):
                try:
                    self.validate_authz_file(self.__file)
                except Exception, e:
                    rcs.restore(self.__file)
                    raise Exception, e
                else:
                    rcs.backup(self.__file, comment=self.comment, user=self.login_as)

        self.comment = []
    
    def validate_authz_file(self, filename):
        if isinstance(filename, basestring):
            stat = os.stat(filename)
            if stat.st_size == 0:
                raise Exception, "Size of file (%s) is zero!" % filename
            
            try:
                from svn import repos as _repos
                _repos.authz_read(filename, 1)
            except ImportError:
                pass
    
    def __str__(self):
        buff = u""
        buff += self.compose_version()
        buff += self.compose_acl()
        buff += u'\n'
        buff += unicode(self.__grouplist)
        buff += u'\n'
        buff += unicode(self.__aliaslist)
        buff += u'\n'
        buff += unicode(self.__reposlist)
        return buff

    def differ(self):
        if isinstance(self.__file, basestring):
            contents_old = unicode(open(self.__file, 'r').read().strip(), "utf-8", "ignore")
        elif isinstance(self.__file, (file, StringIO.StringIO)):
            self.__file.seek(0)
            contents_old = unicode(self.__file.read().strip(), "utf-8", "ignore")
            self.__file.seek(0)
        else:
            return u''
        contents_new = unicode(self).strip()
        if  contents_new != contents_old:
            import difflib
            difflines = []
            for line in difflib.unified_diff(contents_old.splitlines(), contents_new.splitlines(), 'old authz', 'new authz', lineterm=''):
                difflines.append(line)
                if len(difflines)>10:
                    break
            return u'\n'.join(difflines)
        return u''
 
    def parse_groups(self, groups):
        for (name, members) in groups.items():
            group = self.__grouplist.get_or_set(name)
            self.add_group_member(group, members, autodrop=True)

    def parse_module(self, section, contents):
        section = section.strip()

        if ':' in section:
            reposname, path = section.split(':', 1)
        else:
            reposname = '/'
            path = section

        reposname = normalize_repos(reposname)
        path = normalize_path(path)

        self.add_rules(reposname, path, contents, force=True)

    def parse_aliases(self, aliases):
        for (name, username) in aliases.items():
            name = normalize_user(name)
            username = normalize_user(username)
            self.add_alias(name, username)

    def parse_acl(self):
        acls = self.config.initial_comment
        pattern = re.compile(r'^#\s*admin\s*:\s*(.+?)\s*=\s*(.+?)\s*(?:#.*)?$')
        if acls:
            for acl in acls:
                i = pattern.search(acl)
                if i:
                    name  = normalize_repos(i.group(1))
                    admin = normalize_user(i.group(2))
                    if name and admin:
                        repos = self.__reposlist.get_or_set(name)
                        self.set_admin(admin, repos)

    def parse_version(self):
        ic = self.config.initial_comment
        pattern = re.compile(r'^#\s*version\s*[:=]\s*(.+?)\s*(?:#.*)?$')
        if ic:
            for line in ic:
                i = pattern.search(line)
                if i:
                    version  = i.group(1)
                    if version:
                        self.__version = version
                        break

    def get_revision_from_file(self):
        if not self.__file:
            return ''
        else:
            assert isinstance(self.__file, (basestring, file, StringIO.StringIO))
            if isinstance(self.__file, basestring):
                f = open(self.__file)
            else:
                f = self.__file
                f.seek(0,0)
            pattern = re.compile(r'^#\s*version\s*[:=]\s*(.+?)\s*(?:#.*)?$')
            version = ""
            for line in f:
                i = pattern.search(line)
                if i:
                    version  = i.group(1)
                    break
            if isinstance(self.__file, (basestring)):
                f.close()
            else:
                f.seek(0,0)
            return version

    def compose_version(self):
        buff = ""
        if self.__version:
            buff = u"# version : %s\n" % self.__version
        return buff

    def compose_acl(self):
        buff = u""
        for repos in self.__reposlist:
            admins = repos.admins
            if admins:
                buff += u"# admin : %s = %s\n" % (repos.name, admins)
        return buff

    def is_admin(self, user, repos='/', admins=None):
        if isinstance(user, User):
            user = user.uname
        elif isinstance(user, Alias):
            user = user.username
        elif isinstance(user, basestring):
            user = normalize_user(user)
        elif not user:
            return False

        if isinstance(repos, basestring):
            repos = self.__reposlist.get(repos)
        
        if repos:
            if admins is None:
                admins = repos.admins
            for i in admins.split(','):
                if i: i = normalize_user(i)
                
                if not i: continue
                
                if user == i:
                    return True

                i = self.get_userobj(i, autocreate=False)

                if i and user in i:
                    return True

        if not repos or repos.name != '/':
            return self.is_admin(user, '/')
        else:
            return False

    def is_super_user(self, user):
        return self.is_admin(user, '/')
    
    def set_admin(self, admins, repos=None):
        if not isinstance(repos, Repos):
            repos = self.__reposlist.get(repos)
        if not repos:
            return False
        
        if isinstance(admins, basestring):
            alist = [x.strip() for x in admins.split(',')]
        elif isinstance(admins, (list, tuple, set)):
            alist = admins
        else:
            alist = [admins]
        ulist = []
        for i in alist:
            if isinstance(i, (User, Group, Alias)):
                ulist.append(i)
            elif not i:
                continue
            elif isinstance(i, basestring):
                ulist.append(self.get_userobj(i, autocreate=True))
            else:
                raise Exception, "unknown user: %s, type: %s" % (i, type(i))

        repos.admins = ulist

        return True

    def add_repos(self, reposname):
        repos = self.__reposlist.get_or_set(reposname)
        return repos

    def get_repos(self, reposname):
        repos = self.__reposlist.get(reposname)
        return repos

    def del_repos(self, name, recursive=False):
        return self.__reposlist.remove(name, recursive)

    def add_module(self, reposname, path):
        repos = self.__reposlist.get_or_set(reposname)
        module = repos.add_module(path)
        return module

    def get_module(self, reposname, path):
        repos = self.__reposlist.get(reposname)
        if repos:
            module = repos.get_module(path)
            return module
        else:
            return None

    def del_module(self, reposname, path):
        repos = self.__reposlist.get(reposname)
        if repos:
            return repos.del_module(path)
        return False

    def add_rules(self, reposname, path, rules, force=False):
        module = self.get_module(reposname, path)
        if not module:
            if force:
                module = self.add_module(reposname, path)
            else:
                raise Exception, 'No module exist for %s:%s' % (reposname, path)

        rdict = {}
        if isinstance(rules, (str, unicode)):
            rules = rules.strip()
            if '\n' in rules:
                rule_list = rules.split('\n')
            elif ';' in rules:
                rule_list = rules.split(';')
            else:
                rule_list = [rules,]
            for rule in rule_list:
                rule = rule.strip()
                if not rule:
                    continue
                if '=' in rule:
                    key, value = rule.split('=', 1)
                    rdict[key.strip()] = value.strip()
                else:
                    raise Exception, _('Unknown rule format: %s') % rule
        elif isinstance(rules, (list, tuple)):
            for rule in rules:
                if '=' in rule:
                    key, value = rule.split('=', 1)
                    rdict[key.strip()] = value.strip()
                else:
                    raise Exception, _('Unknown rule format: %s') % rule
        elif isinstance(rules, dict):
            rdict = rules
        else:
            raise Exception, _('Unknown rule format: %s') % rules

        for (name, rights) in rdict.items():
            obj = self.get_userobj(name, autocreate=True)
            module.update_rule(obj, rights)

        return True

    def set_rules(self,reposname, path, rules, reset=True, force=False):
        module = self.get_module(reposname, path)
        if not module:
            #if force:
            #    module = self.add_module(reposname, path)
            #else:
            raise Exception, _('No module exist for %s:%s') % (reposname, path)

        if reset:
            module.clean_rules()

        return self.add_rules(reposname, path, rules, force=force)
 
    def get_userobj(self, name, autocreate=False):
        name = normalize_user(name)

        if not name:
            return None
        if name[0] == '@' or name[0] == '$' or name == '*':
            obj = self.grouplist.get_or_set(name, autocreate = autocreate)
        elif name[0] == '&':
            obj = self.aliaslist.get_or_set(name, autocreate = autocreate)
        else:
            obj = self.userlist.get_or_set(name, autocreate = autocreate)
        return obj

    def get_manageable_repos_list(self, username):
        repos_list = []
        if self.is_admin(username, '/'):
            repos_list.extend(map(lambda x:x.name, self.reposlist))
        else:
            for i in self.reposlist:
                if self.is_admin(username, i):
                    repos_list.append(i.name)
        return sorted(repos_list)
        
    def del_rule(self, reposname, path, rule):
        module = self.get_module(reposname, path)
        if not module:
            return False
        return module.del_rule(rule)

    def add_group(self, name, members=None, autodrop=False):
        group = self.__grouplist.get_or_set(name)
        if members:
            return self.add_group_member(group, members, autodrop=autodrop)
        return group

    def set_group(self, name, members, autodrop=False):
        group = self.__grouplist.get_or_set(name)
        group.remove_all()
        if members:
            return self.add_group_member(group, members, autodrop=autodrop)
        return group

    def del_group(self, name, force=False):
        self.chk_grp_ref_by_rules(name)
        return self.__grouplist.remove(name, force=force)

    def add_group_member(self, group, members, autodrop=False):
        if isinstance(group, basestring):
            groupobj = self.__grouplist.get_or_set(group)
        else:
            groupobj = group

        ulist = []
        if isinstance(members, (str, unicode)):
            for user in members.split(','):
                user = normalize_user(user)
                if user:
                    ulist.append(user)
        elif isinstance(members, (list, tuple)):
            ulist.extend(members)
        else:
            ulist.append(members)

        for user in ulist:
            if isinstance(user, (User, Group, Alias)):
                member = user
            else:
                member = self.get_userobj(user, autocreate=True)
            if member:
                groupobj.append(member, autodrop=autodrop)

        return groupobj

    def del_group_member(self, groupname, username):
        group = self.__grouplist.get_or_set(groupname)
        return group.remove(username)

    def add_user(self, username):
        username = normalize_user(username)
        userobj = self.userlist.get_or_set(username)
        return userobj

    def add_alias(self, aliasname, user=None):
        alias = self.__aliaslist.get_or_set(aliasname)
        if isinstance(user, basestring):
            user = self.userlist.get_or_set(user)
        if user:
            alias.user = user
        return alias

    def del_alias(self, name, force=False):
        alias = self.__aliaslist.get(name)
        if not alias:
            return False

        self.chk_alias_ref_by_rules(alias)

        for group in self.__grouplist:
            if alias in group.memberobjs:
                if force:
                    group.memberobjs.remove(alias)
                else:
                    raise Exception, \
                        _('Alias %s is referenced by group %s.') % \
                            (alias.uname, group.uname)
        return self.__aliaslist.remove(name)

    def __check_ref_by_rules(self, name):
        if name:
            if isinstance(name, (User, Group, Alias)):
                name = name.uname

            for i in self.modulelist():
                ulist = map(lambda x:x.uname, i)
                if name in ulist:
                    raise Exception, _("%s is referenced by [%s].") % \
                            (name, i.fullname)

    def chk_alias_ref_by_rules(self, name):
        if isinstance(name, (str, unicode)):
            if name and name[0] != '&':
                name = '&'+name
        return self.__check_ref_by_rules(name)

    def chk_grp_ref_by_rules(self, name):
        if isinstance(name, (str, unicode)):
            if name and name[0] != '@' and name[0] != '$' and name != '*' :
                name = '@'+name
        return self.__check_ref_by_rules(name)

    def check_rights(self, user, repos, path, required):
        if isinstance(user, (str, unicode)):
            user = normalize_user(user)
        if isinstance(repos, (str, unicode)):
            repos = normalize_repos(repos)
        if isinstance(path, (str, unicode)):
            path  = normalize_path(path)

        if isinstance(required, (str, unicode)):
            required = required.strip()
        else:
            required = int(required) & RIGHTS_ALL

        if not user:
            user = '*'
        if not repos:
            repos = '/'
        if not path:
            path  = '/'

        module = self.get_module(repos, path)
        if module:
            if module.access_is_determined(user):
                return module.access_is_granted(user, required)

        if repos != '/':
            module = self.get_module('/', path)
            if module:
                if module.access_is_determined(user):
                    return module.access_is_granted(user, required)

        if '/' in path and path != '/':
            path = path.rsplit('/', 1)[0]
            return self.check_rights(user, repos, path, required)
        else:
            return False

    def get_rights(self, user, repos, path):
        user = normalize_user(user)
        repos = normalize_repos(repos)
        path = normalize_path(path)

        module = self.get_module(repos, path)
        if module:
            if module.access_is_determined(user):
                return module.get_permit_bits(user)

        if repos != '/':
            module = self.get_module('/', path)
            if module:
                if module.access_is_determined(user):
                    return module.get_permit_bits(user)

        if '/' in path and path != '/':
            path = path.rsplit('/', 1)[0]
            return self.get_rights(user, repos, path)
        else:
            return (RIGHTS_NONE, RIGHTS_ALL)

    def get_access_map(self, user=None, reposname='/', descend=True):
        if isinstance(user, (str, unicode)):
            user = normalize_user(user)
        if not user:
            user = '*'
        reposname = normalize_repos(reposname)
        repos = self.get_repos(reposname)
        if not repos:
            if descend and reposname != '/':
                return self.get_access_map(user, '/')
            return None

        path_set = set(map(lambda x:x.path, repos.module_list))
        read_path_set = set()
        write_path_set = set()
        deny_path_set = set()

        if reposname != '/':
            root_path_set = set(map(lambda x:x.path, 
                                    self.get_repos('/').module_list))
            path_set = path_set.union(root_path_set)

        for path in path_set:
            perm, deny = self.get_rights(user, reposname, path)
            if perm & RIGHTS_W:
                write_path_set.add(path)
            elif perm & RIGHTS_R:
                read_path_set.add(path)
            else:
                deny_path_set.add(path)
        return { 
                'write':write_path_set, 
                'read': read_path_set, 
                'deny': deny_path_set }

    def get_path_access_msgs(self, user, reposname, path, abbr=False):
        if isinstance(user, (str, unicode)):
            user = normalize_user(user)
        if not user:
            user = '*'
        if reposname:
            reposname = normalize_repos(reposname)
        if path:
            path = normalize_path(path)
        msgs = []
        if not reposname or reposname=='*' or reposname=='...':
            reposname = sorted(map(lambda x: x.name, self.__reposlist))
        
        if isinstance(reposname, (list, tuple)):
            for i in reposname:
                msgs.extend(self.get_path_access_msgs(user, i, path, abbr))
        else:
            perm, deny = self.get_rights(user, reposname, path)
            if abbr:
                tmpstr = '[%s:%s] %s =' % ( reposname, path, user)
                if perm & RIGHTS_W:
                    tmpstr += ' rw'
                elif perm & RIGHTS_R:
                    tmpstr += ' r'
                msgs.append(tmpstr)
            else:
                if perm & RIGHTS_W:
                    msgs.append(_('User %(username)s has Full (RW) rights for module %(repos)s:%(path)s') % {'username':user, 'repos':reposname, 'path':path})
                elif perm & RIGHTS_R:
                    msgs.append(_('User %(username)s has ReadOnly (RO) rights for module %(repos)s:%(path)s') % {'username':user, 'repos':reposname, 'path':path})
                else:
                    msgs.append(_('User %(username)s can *NOT* access to module %(repos)s:%(path)s') % {'username':user, 'repos':reposname, 'path':path})
            
        return msgs

    def get_access_map_msgs(self, user='*', reposname=None, abbr=False):
        if isinstance(user, (str, unicode)):
            user = normalize_user(user)
        if not user:
            user = '*'

        if isinstance(reposname, basestring):
            reposname = normalize_repos(reposname)

        maps = []
        msgs = []
        if not reposname or reposname=='*' or reposname=='...':
            reposname = [r.name for r in sorted(self.__reposlist)]
        
        if isinstance(reposname, (list, tuple)):
            for i in reposname:
                #map = self.get_access_map(user, i, descend=False)
                map = self.get_access_map(user, i)
                log.debug("repos:%s, map: %s" % (i, map))
                if map:
                    map['user'] = unicode(user)
                    map['repos'] = i
                    maps.append(map)
        else:
            map = self.get_access_map(user, reposname)
            if map: 
                map['user'] = unicode(user)
                map['repos'] = reposname
                maps.append(map)

        for map in maps:
            if abbr:
                write_list = ', '.join(sorted(map['write']))
                read_list  = ', '.join(sorted(map['read']))
                deny_list  = ', '.join(sorted(map['deny']))

                msg = _('''
%(user)s => [%(repos)s]
%(sep)s
RW: %(write)s
RO: %(read)s
XX: %(deny)s
\n''') %  { 
                 'repos' : map['repos'], 
                 'user'  : map['user'], 
                 'sep'   : '-'*40, 
                 'write' : write_list, 
                 'read'  : read_list, 
                 'deny'  : deny_list, 
                }
            else:
                write_list = '    '+'\n    '.join(sorted(map['write']))
                read_list  = '    '+'\n    '.join(sorted(map['read']))
                deny_list  = '    '+'\n    '.join(sorted(map['deny']))

                msg = _('''
%(heading)s
Access map on '%(repos)s' for user '%(user)s'
%(heading)s
  * Writable:
%(write)s
%(sep)s
  * Readable:
%(read)s
%(sep)s
  * Denied:
%(deny)s
%(sep)s\n''') %  { 
                 'repos' : map['repos'], 
                 'user'  : map['user'], 
                 'heading' : '='*50, 
                 'sep'   : '-'*40, 
                 'write' : write_list, 
                 'read'  : read_list, 
                 'deny'  : deny_list, 
                }
            msgs.append(msg)
        return msgs


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008 OpenSourceXpress Ltd. (http://www.ossxp.com)
# Author: Jiang Xin
# Contact: http://www.ossxp.com
#          http://www.worldhello.net
#          http://moinmo.in/JiangXin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import logging

from pysvnmanager.lib.base import *
from pysvnmanager.lib.text import to_unicode
from pysvnmanager.model.svnauthz import *
from pysvnmanager.model import repos as _repos
from pylons.i18n import _, ungettext, N_

log = logging.getLogger(__name__)

class AuthzController(BaseController):
    requires_auth = True

    def __init__(self):
        c.menu_active = "authz"
        try:
            self.authz = SvnAuthz(cfg.authz_file)
            self.login_as = session.get('user')
            # Used as checked in user to rcs file.
            self.authz.login_as = self.login_as
            self.own_reposlist = set(self.authz.get_manageable_repos_list(self.login_as))
            
            # self.reposlist_new is what in ReposRoot directory.
            self.all_reposlist = set(_repos.Repos(cfg.repos_root).repos_list)
            self.all_reposlist.add('/')
            
            self.reposlist_set = self.own_reposlist & self.all_reposlist
            self.reposlist_unexist = self.own_reposlist - self.all_reposlist
            
            self.is_super_user = self.authz.is_super_user(self.login_as)
            if self.is_super_user:
                self.reposlist_unset = self.all_reposlist - self.own_reposlist
            else:
                self.reposlist_unset = set()
        except Exception, e:
            import traceback
            g.catch_e = [unicode(e), traceback.format_exc(5) ]
            return
            
    def __before__(self, action=None):
        super(AuthzController, self).__before__(action)
        if not self.own_reposlist and not self.is_super_user:
            return redirect(url(controller='security', action='failed'))
        diff = self.authz.differ()
        if diff:
            c.global_message = _('Some one maybe you, has modified the svn authz file by hands. Please save once to fix possible config error.') + "<blockquote>" + "<br>".join(diff.splitlines()) + "</blockquote>"

    def index(self):
        c.revision = self.authz.version
        c.is_super_user = self.is_super_user
        # used for functional test.
        c.reposlist = self.own_reposlist
        all_avail_users = []
        all_avail_users.append([_("All users(with anon)"), '*'])
        all_avail_users.append([_("Known users"), '$authenticated'])
        all_avail_users.append([_("Anonymous"), '$anonymous'])
        for group in self.authz.grouplist:
            i = group.uname
            if i == '*' or i =='$authenticated' or i == '$anonymous':
                continue
            all_avail_users.append([_("Group:")+i[1:], i])
        for alias in self.authz.aliaslist:
            i = alias.uname
            all_avail_users.append([_("Alias:")+i[1:], i])
        for user in self.authz.userlist:
            i = user.uname
            all_avail_users.append([i, i])
        
        c.all_avail_users = all_avail_users

        return render('/authz/index.mako')

    def init_repos_list(self):
        total = 0;
        msg = ''
 
        # get javascript code for top_form's role_selector
        msg += 'id[0]="%s";' % '...'
        msg += 'name[0]="%s";\n' % _("Please choose...")
        total += 1;
        for reposname in sorted(self.reposlist_set):
            msg += 'id[%d]="%s";' % (total, reposname)
            msg += 'name[%d]="%s";\n' % (total, reposname)
            total += 1;
        for reposname in sorted(self.reposlist_unexist):
            msg += 'id[%d]="%s";' % (total, reposname)
            msg += 'name[%d]="%s";\n' % (total, reposname+' (?)')
            total += 1;
        for reposname in sorted(self.reposlist_unset):
            msg += 'id[%d]="%s";' % (total, reposname)
            msg += 'name[%d]="%s";\n' % (total, reposname+' (!)')
            total += 1;
        msg += 'total=%d;\n' % total
        msg += 'revision="%s";\n' % self.authz.version
        return msg

    def repos_changed(self):
        total = 0;
        msg = ''
        d = request.params
        select = d.get('select')
        repos = self.authz.get_repos(select)
        if not repos:
            log.warning("Repos '%s' not exists. Create authz config automatically." % select)
            repos = self.authz.add_repos(select)

        if repos:
            # get javascript code for top_form's role_selector
            msg += 'id[0]="%s";' % '...'
            msg += 'name[0]="%s";\n' % _("Please choose...")
            total += 1;
            for path in repos.path_list:
                msg += 'id[%d]="%s";' % (total, path)
                msg += 'name[%d]="%s";\n' % (total, path)
                total += 1;
            msg += 'total=%d;\n' % total
            msg += 'admin_users="%s";\n' % repos.admins
            msg += 'revision="%s";\n' % self.authz.version
        
        return msg

    def path_changed(self):
        total = 0;
        msg = ''
 
        d = request.params
        reposname = d.get('reposname')
        path = d.get('path')
        module = self.authz.get_module(reposname, path)
        if not module:
            return msg;

        # get javascript code for top_form's role_selector
        for rule in module:
            rule = unicode(rule)
            tmp = rule.split('=')
            msg += 'user[%d]="%s";\n' % (total, tmp[0].strip())
            msg += 'rights[%d]="%s";\n' % (total, tmp[1].strip())
            total += 1;
        msg += 'total=%d;\n' % total
        msg += 'revision="%s";\n' % self.authz.version

        return msg
        
    def save_authz(self):
        d = request.params
        
        member_list = []
        msg = ""

        reposname = d.get('reposname')
        admins    = d.get('admins', '')
        path      = d.get('path')
        rules     = d.get('rules')
        revision  = d.get('revision', self.authz.version)
        
        # mode1: new or edit repository
        mode1     = d.get('mode1')
        if mode1 == "new":
            isAddRepos = True
        else:
            isAddRepos = False
        
        # mode2: new or edit module
        mode2     = d.get('mode2')
        if mode2 == "new":
            isAddModule = True
        else:
            isAddModule = False

        log_message = _(u"User %(user)s changed authz rules. (rev:%(rev)s)") % \
                 {'user':session.get('user'), 'rev': revision}

        try:
            if isAddRepos:
                assert self.is_super_user
                repos = self.authz.add_repos(reposname)
            else:
                repos = self.authz.get_repos(reposname)
                if not repos:
                    assert self.is_super_user
                    log.warning("Repos '%s' not exists. Create authz config automatically." % reposname)
                    repos = self.authz.add_repos(reposname)
                
            if not repos:
                raise Exception, _("Repository %s not exist.") % reposname
            
            if path:
                if isAddModule:
                    module = repos.add_module(path)
                else:
                    module = repos.get_module(path)
                if not module:
                    raise Exception, _("Module %s not exist.") % path
            else:
                module = None
            
            if not self.authz.is_admin(self.login_as, repos.name, admins) and \
                not (repos.name != '/' and self.authz.is_super_user(self.login_as)):
                raise Exception, _("You can not delete yourself from admin list.")
            
            self.authz.set_admin(admins, repos)

            if module:
                self.authz.set_rules(reposname, path, rules);
            self.authz.save(revision, comment=log_message)
        except Exception, e:
            msg = to_unicode(e)

        log.info(log_message)
        if msg: log.error(msg)
        
        return msg

    def delete_authz(self):
        d = request.params
        
        member_list = []
        msg = ""

        reposname = d.get('reposname')
        path  = d.get('path')
        revision  = d.get('revision', self.authz.version)
        
        log_message = _(u"User %(user)s delete authz rules. (rev:%(rev)s)") % \
                         {'user':session.get('user'), 'rev': revision}
        try:
            self.authz.del_module(reposname, path);
            self.authz.save(revision, comment=log_message)
        except Exception, e:
            msg = to_unicode(e)
        
        log.info(log_message)
        if msg: log.error(msg)

        return msg




# -*- coding: utf-8 -*-
# Copyright (C) 2008 OpenSourceXpress Ltd. (http://www.ossxp.com)
# Author: Jiang Xin
# Contact: http://www.ossxp.com
#          http://www.worldhello.net
#          http://moinmo.in/JiangXin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

"""The application's Globals object"""

from beaker.cache import CacheManager
from beaker.util import parse_cache_config_options

class Globals(object):
    """Globals acts as a container for objects available throughout the
    life of the application

    """

    def __init__(self, config):
        """One instance of Globals is created during application
        initialization and is available during requests via the
        'app_globals' variable

        """
        self.cache = CacheManager(**parse_cache_config_options(config))



## -*- coding: utf-8 -*-
#
# Copyright (C) 2008 OpenSourceXpress Ltd. (http://www.ossxp.com)
# Author: Jiang Xin
# Contact: http://www.ossxp.com
#          http://www.worldhello.net
#          http://moinmo.in/JiangXin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

from pysvnmanager.tests import *
from pysvnmanager.controllers import authz

class TestAuthzController(TestController):
    def test_index(self):
        # Test redirect to login pange
        res = self.app.get(url(controller='authz', action='index'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/login'), res

        # Login as common user
        self.login('nobody')
        res = self.app.get(url(controller='authz', action='index'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/security/failed'), res.header
        
        # Login as repos admin
        self.login('admin2')
        res = self.app.get(url(controller='authz', action='index'))
        assert res.status == "200 OK", res.status
        assert ','.join(sorted(res.tmpl_context.reposlist)) == 'repos1,repos2', res.tmpl_context.reposlist
        assert """<input type="button" name="save_btn"   value='Save'""" in res.body, res.body

        # Login as superuser
        self.login('root')
        res = self.app.get(url(controller='authz', action='index'))
        assert res.status == "200 OK", res.status
        assert ','.join(sorted(res.tmpl_context.reposlist)) == u'/,document,repos1,repos2,repos3', res.tmpl_context.reposlist
        assert """<input type="button" name="save_btn"   value='Save'""" in res.body, res.body

    def test_init_repos_list(self):
        # authn test
        res = self.app.get(url(controller='authz', action='init_repos_list'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/login'), res.location

        # authz test
        self.login('nobody')
        res = self.app.get(url(controller='authz', action='init_repos_list'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/security/failed'), res.location

        # Login as superuser
        self.login('root')
        res = self.app.get(url(controller='authz', action='init_repos_list'))
        assert res.status == "200 OK", res.status
        assert """id[0]="...";name[0]="Please choose...";
id[1]="/";name[1]="/";
id[2]="repos3";name[2]="repos3";
id[3]="document";name[3]="document (?)";
id[4]="repos1";name[4]="repos1 (?)";
id[5]="repos2";name[5]="repos2 (?)";
id[6]="project1";name[6]="project1 (!)";
id[7]="project2";name[7]="project2 (!)";
total=8;
revision="0.2.1";
""" == res.body, res.body
    
    def test_repos_changed(self):
        # authn test
        res = self.app.get(url(controller='authz', action='repos_changed'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/login'), res.location

        # authz test
        self.login('nobody')
        res = self.app.get(url(controller='authz', action='repos_changed'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/security/failed'), res.location

        # Login as superuser
        self.login('root')
        params = {'select':'/',}
        res = self.app.get(url(controller='authz', action='repos_changed'), params)
        assert res.status == "200 OK", res.status
        assert '''id[0]="...";name[0]="Please choose...";
id[1]="/trunk/src";name[1]="/trunk/src";
id[2]="/trunk";name[2]="/trunk";
id[3]="/";name[3]="/";
id[4]="/tags";name[4]="/tags";
id[5]="/branches";name[5]="/branches";
total=6;
admin_users="&admin, root";
revision="0.2.1";
''' == res.body, res.body

        params = {'select':'repos1',}
        res = self.app.get(url(controller='authz', action='repos_changed'), params)
        assert res.status == "200 OK", res.status
        assert '''id[0]="...";name[0]="Please choose...";
id[1]="/trunk/src";name[1]="/trunk/src";
id[2]="/trunk";name[2]="/trunk";
id[3]="/";name[3]="/";
total=4;
admin_users="@admin";
revision="0.2.1";
''' == res.body, res.body

    def test_path_changed(self):
        params={}
        # authn test
        res = self.app.get(url(controller='authz', action='path_changed'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/login'), res.location

        # authz test
        self.login('nobody')
        res = self.app.get(url(controller='authz', action='path_changed'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/security/failed'), res.location

        self.login('root')
        params = {'reposname':'/', 'path':u'/tags//'}
        res = self.app.get(url(controller='authz', action='path_changed'), params)
        assert res.status == "200 OK", res.status
        assert '''user[0]="&pm";
rights[0]="rw";
user[1]="$authenticated";
rights[1]="r";
total=2;
revision="0.2.1";
''' == res.body, res.body

        self.login('root')
        params = {'reposname':'document', 'path':'/trunk/商务部'}
        res = self.app.get(url(controller='authz', action='path_changed'), params)
        assert res.status == "200 OK", res.status
        assert '''user[0]="*";
rights[0]="";
user[1]="@admin";
rights[1]="rw";
user[2]="@biz";
rights[2]="rw";
total=3;
revision="0.2.1";
''' == res.body, res.body

        self.login('root')
        params = {'reposname':'/', 'path':'/noexist'}
        res = self.app.get(url(controller='authz', action='path_changed'), params)
        assert res.status == "200 OK", res.status
        assert '' == res.body, res.body


    def test_set_repos_admin(self):
        # authn test
        try:
            res = self.app.get(url(controller='authz', action='save_authz'))
            assert res.status == "302 Found", res.status
            assert res.location.endswith('/login'), res.location

            # authz test
            self.login('nobody')
            res = self.app.get(url(controller='authz', action='save_authz'))
            assert res.status == "302 Found", res.status
            assert res.location.endswith('/security/failed'), res.location
            
            # Login as superuser
            self.login('root')
            params = {'reposname':'/', 'admins':''}
            res = self.app.get(url(controller='authz', action='save_authz'), params)
            assert res.status == "200 OK", res.status
            assert "You can not delete yourself from admin list." == res.body, res.body
        finally:
            self.rollback()
            
        try:
            params = {'reposname':'/', 'admins':'root, @some'}
            res = self.app.get(url(controller='authz', action='save_authz'), params)
            assert res.status == "200 OK", res.status
            assert "" == res.body, res.body
        finally:
            self.rollback()

        try:
            self.login('jiangxin')
            params = {'reposname':'/', 'admins':'&admin'}
            res = self.app.get(url(controller='authz', action='save_authz'), params)
            assert res.status == "200 OK", res.status
            assert "" == res.body, res.body
        finally:
            self.rollback()
            
        try:
            self.login('jiangxin')
            params = {'reposname':'/', 'admins':'root'}
            res = self.app.get(url(controller='authz', action='save_authz'), params)
            assert res.status == "200 OK", res.status
            assert "You can not delete yourself from admin list." == res.body, res.body
        finally:
            self.rollback()

        try:
            self.login('root')
            params = {'reposname':'/repos1', 'admins':'user1'}
            res = self.app.get(url(controller='authz', action='save_authz'), params)
            assert res.status == "200 OK", res.status
            assert "" == res.body, res.body
        finally:
            self.rollback()
        
        try:
            self.login('root')
            params = {'reposname':'/repos1', 'admins':'user1, root'}
            res = self.app.get(url(controller='authz', action='save_authz'), params)
            assert res.status == "200 OK", res.status
            assert "" == res.body, res.body
        finally:
            self.rollback()
        
        try:
            self.login('admin1')
            params = {'reposname':'/repos1', 'admins':'user1, root'}
            res = self.app.get(url(controller='authz', action='save_authz'), params)
            assert res.status == "200 OK", res.status
            assert "You can not delete yourself from admin list." == res.body, res.body

            self.login('admin1')
            params = {'reposname':'/repos1', 'admins':'admin1, admin2'}
            res = self.app.get(url(controller='authz', action='save_authz'), params)
            assert res.status == "200 OK", res.status
            assert "" == res.body, res.body
        finally:
            self.rollback()

    def test_set_rules(self):
        # Modify rules for repos
        try:
            authz = self.load_authz()
            module1 = authz.get_module('repos1', u'trunk/src')
            self.assert_(module1 != None, type(module1))
            self.assert_(unicode(module1)=='[repos1:/trunk/src]\nuser1 = \n', unicode(module1).encode('utf-8'))

            self.login('root')
            params = {'reposname':'/repos1', 'path':'/trunk/src', 'admins':'蒋鑫', 'rules':'@管理员=rw\n&别名1=r\n*=\nuser2=r', 'mode1':'edit', 'mode2':'edit' }
            res = self.app.get(url(controller='authz', action='save_authz'), params)
            assert res.status == "200 OK", res.status
            assert "" == res.body, res.body
            
            authz = self.load_authz()
            module1 = authz.get_module('repos1', u'trunk/src')
            self.assert_(module1 != None, type(module1))
            self.assert_(unicode(module1)==u'[repos1:/trunk/src]\n&别名1 = r\n* = \n@管理员 = rw\nuser2 = r\n', unicode(module1).encode('utf-8'))
            
            # Test login using chinese username
            self.login('蒋鑫')
            params = {'reposname':'/repos1', 'path':'/trunk/src', 'admins':'其他', 'rules':'@管理员=rw\n&别名1=r\n*=\nuser2=r', 'mode1':'edit', 'mode2':'edit' }
            res = self.app.get(url(controller='authz', action='save_authz'), params)
            assert res.status == "200 OK", res.headers
            assert "You can not delete yourself from admin list." in res.body, res.body
        finally:
            self.rollback()

        # Add New Repos
        try:
            authz = self.load_authz()
            repos1 = authz.get_repos('reposX')
            self.assert_(repos1 == None, type(repos1))

            self.login('root')
            params = {'reposname':'reposX', 'admins':'蒋鑫', 'rules':'@管理员=rw\n&别名1=r\n*=\nuser2=r', 'mode1':'new', 'mode2':'new' }
            res = self.app.get(url(controller='authz', action='save_authz'), params)
            assert res.status == "200 OK", res.status
            assert "" == res.body, res.body
        
            authz = self.load_authz()
            repos1 = authz.get_repos('reposX')
            self.assert_(repos1 != None, type(repos1))
            self.assert_(unicode(repos1)==u'', unicode(repos1).encode('utf-8'))
            self.assert_(repos1.admins==u'蒋鑫', repos1.admins.encode('utf-8'))
        finally:
            self.rollback()

        # Add New Repos with Module/Rules
        try:
            authz = self.load_authz()
            repos1 = authz.get_repos('reposX')
            self.assert_(repos1 == None, type(repos1))

            self.login('root')
            params = {'reposname':'reposX', 'admins':'蒋鑫', 'path':'/项目a', 'rules':'@管理员=rw\n&别名1=r\n*=\nuser2=r', 'mode1':'new', 'mode2':'new' }
            res = self.app.get(url(controller='authz', action='save_authz'), params)
            assert res.status == "200 OK", res.status
            assert "" == res.body, res.body
        
            authz = self.load_authz()
            repos1 = authz.get_repos('reposX')
            self.assert_(unicode(repos1)==u'[reposX:/项目a]\n&别名1 = r\n* = \n@管理员 = rw\nuser2 = r\n\n', unicode(repos1).encode('utf-8'))
            self.assert_(repos1.admins==u'蒋鑫', repos1.admins.encode('utf-8'))
        finally:
            self.rollback()
        
        # Test Repos/Module not exist Exception
        try:
            self.login('root')
            params = {'reposname':'reposX', 'path':'/trunk/src', 'admins':'蒋鑫', 'rules':'@管理员=rw\n&别名1=r\n*=\nuser2=r', 'mode1':'edit', 'mode2':'edit' }
            res = self.app.get(url(controller='authz', action='save_authz'), params)
            assert res.status == "200 OK", res.status
            assert "Module /trunk/src not exist." == res.body, res.body
        finally:
            self.rollback()
        
        # Test Repos/Module not exist Exception
        try:
            self.login('root')
            params = {'reposname':'repos1', 'path':'/trunk/myproject', 'admins':'蒋鑫', 'rules':'@管理员=rw\n&别名1=r\n*=\nuser2=r', 'mode1':'edit', 'mode2':'edit' }
            res = self.app.get(url(controller='authz', action='save_authz'), params)
            assert res.status == "200 OK", res.status
            assert "Module /trunk/myproject not exist." == res.body, res.body
        finally:
            self.rollback()
        
        
    def test_delete_authz(self):
        # authn test
        res = self.app.get(url(controller='authz', action='delete_authz'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/login'), res.location

        # authz test
        self.login('nobody')
        res = self.app.get(url(controller='authz', action='delete_authz'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/security/failed'), res.location

        authz = self.load_authz()
        module1 = authz.get_module('document', u'/trunk/行政部')
        self.assert_(module1 != None, type(module1))
 
        self.login('root')
        params = {'reposname':'document', 'path':'/trunk/行政部'}
        res = self.app.get(url(controller='authz', action='delete_authz'), params)

        authz = self.load_authz()
        module1 = authz.get_module('document', u'/trunk/行政部')
        self.assert_(module1 == None, type(module1))        
        
        try:
            self.login('root')
            params = {'reposname':'document', 'path':'/trunk/行政部', 'revision':'123'}
            res = self.app.get(url(controller='authz', action='delete_authz'), params)
            assert res.status == "200 OK", res.status
            assert "Update failed! You are working on a out-of-date revision." in res.body, res.body
        finally:
            self.rollback()



## -*- coding: utf-8 -*-
#
# Copyright (C) 2008 OpenSourceXpress Ltd. (http://www.ossxp.com)
# Author: Jiang Xin
# Contact: http://www.ossxp.com
#          http://www.worldhello.net
#          http://moinmo.in/JiangXin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

from pysvnmanager.tests import *
from pysvnmanager.controllers import check

class TestCheckController(TestController):

    def test_index(self):
        # Test redirect to login pange
        res = self.app.get(url(controller='check', action='index'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/login'), res.location

        # Login as common user
        self.login('nobody')
        res = self.app.get(url(controller='check', action='index'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/security/failed'), res.location
        
        # Login as repos admin
        self.login('admin1')
        res = self.app.get(url(controller='check', action='index'))
        assert res.tmpl_context.reposlist == [u'repos1'], res.tmpl_context.reposlist

        # Login as repos admin
        self.login('admin2')
        res = self.app.get(url(controller='check', action='index'))
        assert res.tmpl_context.reposlist == [u'repos1', u'repos2'], res.tmpl_context.reposlist

        # Login as superuser
        self.login('root')
        res = self.app.get(url(controller='check', action='index'))
        assert res.status == "200 OK", res.status
        assert '''<input type="submit" name="submit" value='Check Permissions'>''' in res.body, res.body
        assert res.tmpl_context.reposlist == [u'/', u'document', u'project1', u'project2', u'repos1', u'repos2', u'repos3'], res.tmpl_context.reposlist


    def test_access_map(self):
        # authn test
        res = self.app.get(url(controller='check', action='access_map'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/login'), res.location

        # authz test
        self.login('nobody')
        res = self.app.get(url(controller='check', action='access_map'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/security/failed'), res.location
        
        # Login as superuser
        self.login('root')
        params = {
                  'userinput':'select', 
                  'userselector':'user1',
                  'reposinput':'select', 
                  'reposselector':'///repos1',
                  'pathinput':'select',
                  'pathselector':'/trunk/src/test',
                  'abbr':'True',
                  }
        res = self.app.get(url(controller='check', action='access_map'), params)
        assert res.status == "200 OK", res.status
        assert '''<div id='acl_path_msg'>[repos1:/trunk/src/test] user1 =</div>''' in res.body, res.body

        params = {
                  'userinput':'select', 
                  'reposinput':'select', 
                  'pathinput':'select',
                  'userselector':'user1',
                  'reposselector':'reposX',
                  'pathselector':'/trunk/src/test',
                  'abbr':'False',
                  }
        res = self.app.get(url(controller='check', action='access_map'), params)
        assert '''<div id='acl_path_msg'>User user1 has ReadOnly (RO) rights for module reposX:/trunk/src/test</div><pre>
==================================================
Access map on 'reposX' for user 'user1'
==================================================
  * Writable:
    
----------------------------------------
  * Readable:
    /branches
    /tags
    /trunk/src
----------------------------------------
  * Denied:
    /
    /trunk
----------------------------------------
''' in res.body, res.body

        params = {
                  'userinput':'manual', 
                  'reposinput':'manual', 
                  'pathinput':'manual',
                  'username':'user2',
                  'reposname':'repos1',
                  'pathname':'/trunk/src/test',
                  'abbr':'1',
                  }
        res = self.app.get(url(controller='check', action='access_map'), params)
        assert '''<div id='acl_path_msg'>[repos1:/trunk/src/test] user2 = r</div>''' in res.body, res.body

        params = {
                  'userinput':'select', 
                  'reposinput':'select', 
                  'pathinput':'manual',
                  'userselector':'user2',
                  'reposselector':'reposX',
                  'pathname':'/trunk/',
                  'abbr':'1',
                  }
        res = self.app.get(url(controller='check', action='access_map'), params)
        assert '''<div id='acl_path_msg'>[reposX:/trunk] user2 =</div>''' in res.body, res.body

        params = {
                  'userinput':'select', 
                  'reposinput':'select', 
                  'pathinput':'select',
                  'userselector':'user3',
                  'reposselector':'repos1',
                  'pathselector':'/trunk',
                  'abbr':'1',
                  }
        res = self.app.get(url(controller='check', action='access_map'), params)
        assert '''<div id='acl_path_msg'>[repos1:/trunk] user3 =</div>''' in res.body, res.body

        params = {
                  'userinput':'select', 
                  'reposinput':'select', 
                  'pathinput':'select',
                  'userselector':'user4',
                  'reposselector':'repos1',
                  'pathselector':'/trunk',
                  'abbr':'1',
                  }
        res = self.app.get(url(controller='check', action='access_map'), params)
        assert '''<div id='acl_path_msg'>[repos1:/trunk] user4 = r</div>''' in res.body, res.body

        params = {
                  'userinput':'select', 
                  'reposinput':'select', 
                  'pathinput':'select',
                  'userselector':'user4',
                  'reposselector':'reposX',
                  'pathselector':'/trunk',
                  'abbr':'1',
                  }
        res = self.app.get(url(controller='check', action='access_map'), params)
        assert '''<div id='acl_path_msg'>[reposX:/trunk] user4 = r</div>''' in res.body, res.body

        params = {
                  'userinput':'select', 
                  'reposinput':'select', 
                  'pathinput':'select',
                  'userselector':'user5',
                  'reposselector':'reposX',
                  'pathselector':'/trunk',
                  'abbr':'1',
                  }
        res = self.app.get(url(controller='check', action='access_map'), params)
        assert '''<div id='acl_path_msg'>[reposX:/trunk] user5 =</div>''' in res.body, res.body

        # Repos = *
        self.login('admin2')
        params = {
                  'userinput':'select', 
                  'userselector':'user1',
                  'reposinput':'select', 
                  'reposselector':'*',
                  'pathinput':'select',
                  'pathselector':'/trunk/src/test',
                  'abbr':'True',
                  }
        res = self.app.get(url(controller='check', action='access_map'), params)
        assert res.status == "200 OK", res.status
        assert '''<div id='acl_path_msg'>[repos1:/trunk/src/test] user1 =<br>
[repos2:/trunk/src/test] user1 = r</div><pre>
user1 => [repos1]
----------------------------------------
RW: 
RO: /branches, /tags, /trunk
XX: /, /trunk/src



user1 => [repos2]
----------------------------------------
RW: /, /trunk
RO: /branches, /tags, /trunk/src
XX: 

</pre>''' in res.body, repr(res.body)

        # permision deny test
        self.login('admin2')
        params = {
                  'userinput':'select', 
                  'reposinput':'select', 
                  'pathinput':'select',
                  'userselector':'user1',
                  'reposselector':'repos3',
                  'pathselector':'/trunk/src/test',
                  'abbr':'True',
                  }
        res = self.app.get(url(controller='check', action='access_map'), params)
        assert res.status == "200 OK", res.status
        assert res.body== 'Permission denied.', res.location

    def test_authz_path(self):
        # authn test
        res = self.app.get(url(controller='check', action='get_auth_path'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/login'), res.location

        # authz test
        self.login('nobody')
        res = self.app.get(url(controller='check', action='get_auth_path'))
        assert res.status == "302 Found", res.status
        assert res.location.endswith('/security/failed'), res.location

        self.login('root')
        params = {}
        params['repos'] = '/'
        res = self.app.get(url(controller='check', action='get_auth_path'), params)
        assert res.status == "200 OK", res.status
        assert '''id[0]="...";name[0]="Please choose...";
id[1]="/trunk/src";name[1]="/trunk/src";
id[2]="/trunk";name[2]="/trunk";
id[3]="/";name[3]="/";
id[4]="/tags";name[4]="/tags";
id[5]="/branches";name[5]="/branches";
total=6;
''' == res.body, res.body

        params['repos'] = 'noexist'
        res = self.app.get(url(controller='check', action='get_auth_path'), params)
        assert res.status == "200 OK", res.status
        assert '' == res.body, res.body

        params['repos'] = 'repos1'
        res = self.app.get(url(controller='check', action='get_auth_path'), params)
        assert res.status == "200 OK", res.status
        assert '''id[0]="...";name[0]="Please choose...";
id[1]="/trunk/src";name[1]="/trunk/src";
id[2]="/trunk";name[2]="/trunk";
id[3]="/";name[3]="/";
total=4;
''' == res.body, res.body

        params['repos'] = 'document'
        res = self.app.get(url(controller='check', action='get_auth_path'), params)
        assert res.status == "200 OK", res.status
        assert u'''id[0]="...";name[0]="Please choose...";
id[1]="/branches";name[1]="/branches";
id[2]="/tags";name[2]="/tags";
id[3]="/trunk/.htgroup";name[3]="/trunk/.htgroup";
id[4]="/trunk/tech";name[4]="/trunk/tech";
id[5]="/trunk/tech/.htaccess";name[5]="/trunk/tech/.htaccess";
id[6]="/trunk/商务部";name[6]="/trunk/商务部";
id[7]="/trunk/商务部/.htaccess";name[7]="/trunk/商务部/.htaccess";
id[8]="/trunk/行政部";name[8]="/trunk/行政部";
id[9]="/trunk/行政部/.htaccess";name[9]="/trunk/行政部/.htaccess";
total=10;
''' == unicode(res.body, 'utf-8'), res.body



# -*- coding: utf-8 -*-

try:
    from crypt import crypt
except ImportError:
    # Native crypt is not supported by Windows, use fcrypt instead.
    from fcrypt import crypt

def htpasswd_login(username, password, config):
    authn_file = getattr(config, 'authn_file', '')
    if isinstance(username, str):
        username = unicode(username, 'utf-8')

    if authn_file:
        pwdfile = open(authn_file)
        for line in pwdfile:
            if line.strip().startswith('#') and ':' not in line:
                continue
            user, pwdhash = line.strip().split(':',1)
            if username == unicode(user,'utf-8'):
                if pwdhash == crypt(password, pwdhash[:2]):
                    return True
                break

    return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2008 OpenSourceXpress Ltd. (http://www.ossxp.com)
# Author: Jiang Xin
# Contact: http://www.ossxp.com
#          http://www.worldhello.net
#          http://moinmo.in/JiangXin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

from pysvnmanager.hooks.plugins import *
from pysvnmanager.hooks.plugins import _

class MantisIntegration(PluginBase):

    # Brief name for this plugin.
    name = _("Mantis bugtracking integration")
    
    # Both description and detail are reStructuredText format. 
    # Reference about reStructuredText: http://docutils.sourceforge.net/docs/user/rst/quickref.html

    # Short description for this plugin.
    description = _("Integration Subversion with Mantis bugtracking. "
                    "If commit-log has proper format (contains bugid), "
                    "it will change bug status and append commint-log "
                    "and code differ as comment of bug status change.")
    
    # Long description for this plugin.
    detail = ""
    
    # Hooks-plugin type: T_START_COMMIT, ..., T_POST_UNLOCK
    type = T_POST_COMMIT
    
    # Plugin config option/value in config ini file.
    key = "mantis_integration"
    value = "yes"
    
    section = 'post_commit'
    
    def enabled(self):
        """
        Return True, if this plugin has been installed.
        Simply call 'has_config()'.
        """
        return self.has_config()
    
    def install_info(self):
        """
        Show configurations if plugin is already installed.
        
        return reStructuredText.
        reST reference: http://docutils.sourceforge.net/docs/user/rst/quickref.html
        """
        return self.description
    
    def install_config_form(self):
        """
        This method will be called to build setup configuration form.
        If this plugin needs parameters, provides form fields here.
        Any html and javascript are welcome.
        """
        return ""
        
    def uninstall(self):
        """
        Uninstall hooks-plugin from repository.
        Simply call 'unset_config()' and 'save()'.
        """
        self.unset_config()
        self.save()
    
    def install(self, params=None):
        """
        Install hooks-plugin from repository.
        Simply call 'set_config()' and 'save()'.
        
        Form fields in setup_config() will pass as params.
        """
        self.set_config()
        self.save()
        
def execute(repospath=""):
    """
    Generate and return a hooks plugin object

    @param request: repos full path
    @rtype: Plugin
    @return: Plugin object
    """
    return MantisIntegration(repospath)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008 OpenSourceXpress Ltd. (http://www.ossxp.com)
# Author: Jiang Xin
# Contact: http://www.ossxp.com
#          http://www.worldhello.net
#          http://moinmo.in/JiangXin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages

setup(
    name='pySvnManager',
    version="0.4.1",
    description='SVN authz web management tools.',
    author='Jiang Xin',
    author_email='jiangxin@ossxp.com',
    url='https://sourceforge.net/projects/pysvnmanager',
    install_requires=[
        "Pylons>=0.10rc1",
        "docutils",
        "Babel",
        "Mako>=0.3.2",
        "WebHelpers>=1.0rc1",
        "Routes>=1.12.1",
        #"python-ldap",
    ],
    setup_requires=["PasteScript>=1.6.3"],
    packages=find_packages(exclude=['ez_setup']),
    include_package_data=True,
    test_suite='nose.collector',
    package_data={'pysvnmanager': ['i18n/*/LC_MESSAGES/*.mo', 'config/*.in', ]},
    message_extractors={'pysvnmanager': [
            ('**.py', 'python', None),
            ('templates/**.mako', 'mako', {'input_encoding': 'utf-8'}),
            ('public/**', 'ignore', None)]},
    zip_safe=False,
    paster_plugins=['PasteScript', 'Pylons'],
    entry_points="""
    [paste.app_factory]
    main = pysvnmanager.config.middleware:make_app

    [paste.app_install]
    main = pylons.util:PylonsInstaller
    """,
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008 OpenSourceXpress Ltd. (http://www.ossxp.com)
# Author: Jiang Xin
# Contact: http://www.ossxp.com
#          http://www.worldhello.net
#          http://moinmo.in/JiangXin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import ConfigParser
import os
import time
import logging

from pysvnmanager.model.rest import reSTify

# i18n works only as pysvnmanager (a pylons app) model.
from pylons import config
if not config.has_key('unittest'):
    from pylons.i18n import _
else:
    _ = lambda x:x
    
log = logging.getLogger(__name__)

def getPackageModules(packagefile):
    """
    Return a list of modules for a package, omitting any modules
    starting with an underscore.
    """
    import os, re
    
    packagedir = os.path.dirname(packagefile)
    pyre = re.compile(r"^([^_].*)\.py$")
    dirlist = os.listdir(packagedir)

    matches = [pyre.match(fn) for fn in dirlist]
    modules = [match.group(1) for match in matches if match]

    modules.sort()
    return modules


def getHandler(name, function="execute"):
    """ return a handler function for a given action or None """
    if function:
        fromlist = [function]
    else:
        fromlist = []
    moduleName = 'pysvnmanager.hooks.plugins.' + name
    module = __import__(moduleName, globals(), {}, fromlist)

    if function:
        # module has the obj for module <moduleName>
        return getattr(module, function)
    else:
        # module now has the toplevel module of <moduleName> (see __import__ docs!)
        components = moduleName.split('.')
        for comp in components[1:]:
            module = getattr(module, comp)
        return module

T_START_COMMIT          = 1
T_PRE_COMMIT            = 2
T_POST_COMMIT           = 3
T_PRE_REVPROP_CHANGE    = 4
T_POST_REVPROP_CHANGE   = 5
T_PRE_LOCK              = 6
T_POST_LOCK             = 7
T_PRE_UNLOCK            = 8
T_POST_UNLOCK           = 9

class PluginBase(object):
    """ Base class for hook plugins
    """
    # Plugin id (will be set automatically after instance initialized)
    id = None # __name__.rsplit('.',1)[-1]

    # Both description and detail are reStructuredText format. 
    # Reference about reStructuredText: http://docutils.sourceforge.net/docs/user/rst/quickref.html
    
    # Brief name for this plugin.
    name = ""
    
    # Short description for this plugin.
    description = ""

    # Long description for this plugin.
    detail = ""

    # Hooks-plugin type: T_START_COMMIT, ..., T_POST_UNLOCK
    type = 0
    
    # Store file timestamp as plugin revision, avoid config confliction.
    # Not edit it manually.
    revision = ""
    
    def __init__(self, repospath):
        # Test if repository is exists.
        self.__repospath = repospath
        if not os.path.exists(self.__repospath):
            raise Exception, _("repos '%s' not exist!") % os.path.basename(self.__repospath)

        # Read configuration from the default config ini file.
        self.__configfile = "%s/conf/hooks.ini" % self.__repospath
        self.__read_config()

    def __read_config(self, force=False):
        # only read config file if out-of-date.
        if force or self.__is_outofdate():
            #log.debug("config is outofdate for '%s'." % self.name)
            timestamp = os.path.getmtime(self.__configfile)

            # ConfigParser has not cleanup method, so we initial it again.
            self.cp = ConfigParser.ConfigParser()
            self.cp.read(self.__configfile)

            # Set timestamp as revision
            self.revision = timestamp            
        # if config file not exist.
        elif not self.revision:
            self.cp = ConfigParser.ConfigParser()
            self.cp.add_section('main')
            self.revision = time.time()
    
    def __is_outofdate(self):
        """
        Test if timestamp of the config file is the same we load it last time.
        """
        if os.path.exists(self.__configfile):
            timestamp = os.path.getmtime(self.__configfile)
            if not self.revision or self.revision != timestamp:
                return True
        # not out-of-date if config file not exist, or timestamp not changed.
        return False
    
    def __cmp__(self, obj):
        assert isinstance(obj, PluginBase)
        if self.type == obj.type:
            return cmp(self.id, obj.id)
        else:
            return cmp(self.type, obj.type)
        
        
    def reload(self, force=True):
        """
        Reload the default config ini file, if out-of-date.
        """
        self.__read_config(force)
    
    def save(self):
        """
        Save config to the default config ini file.
        
        If the file modified time is different with the time last load, Conflict raises.
        """
        # Test if timestamp of the config file is the same we load it last time.
        if self.__is_outofdate():
            raise Exception(_("Conflict: plugin '%s' is modified by others.") % self.name)
        
        # Save to file.
        fp = open(self.__configfile, 'w')
        self.cp.write(fp)
        fp.close()
        
        # Update revision as file timestamp.
        self.revision = os.path.getmtime(self.__configfile)

    def get_config(self, key, default="", section=''):
        """
        Get config from the default config file.
        """
        if not section:
            section = self.section
        if not self.cp.has_section(section):
            self.cp.add_section(section)
        if self.cp.has_option(section, key):
            result = self.cp.get(section, key)
        else:
            result = default
        return result

    def has_config(self, key="", value=""):
        """
        Test if self.key = self.value is in the default config ini file.
        """
        if key == "":
            if not hasattr(self, "key"):
                raise Exception, _("Plugin not fully implemented.")
            else:
                key = self.key

        if value == "" and hasattr(self, "value"):
            value = self.value
        
        if hasattr(self, "section"):
            section = self.section
        else:
            section = 'main'
        
        setting = self.get_config(key, section=section)
        if value:
            return setting == value
        elif setting != "":
            return True
        else:
            return False
        
    def set_config(self, key="", value=""):
        """
        add self.key = self.value to default config ini file.
        """
        if key == "":
            if not hasattr(self, "key"):
                raise Exception, _("Plugin not fully implemented.")
            else:
                key = self.key

        if value == "" and hasattr(self, "value"):
            value = self.value
        
        if hasattr(self, "section"):
            section = self.section
        else:
            section = 'main'
            
        if not self.cp.has_section(section):
            self.cp.add_section(section)
        self.cp.set(section, key, value)

    def unset_config(self, key=""):
        """
        Remove self.key from default config ini file.
        """
        if key == "":
            if not hasattr(self, "key"):
                raise Exception, _("Plugin not fully implemented.")
            else:
                key = self.key
        
        if hasattr(self, "section"):
            section = self.section
        else:
            section = 'main'
        
        if self.cp.has_section(section):
            self.cp.remove_option(section, key)
            # test if section is blank after remove option.
            if not self.cp.options(section):
                self.cp.remove_section(section)
    
    def install_info(self):
        """
        Show configurations if plugin is already installed.
        
        return reStructuredText.
        reST reference: http://docutils.sourceforge.net/docs/user/rst/quickref.html
        """
        return self.description
    
    def show_install_info(self):
        return reSTify(self.install_info())
    
    def get_type(self):
        type = "UNDEFINED"
        if self.type == T_START_COMMIT:
            type = "start-commit"
        elif self.type == T_PRE_COMMIT:
            type = "pre-commit"
        elif self.type == T_POST_COMMIT:
            type = "post-commit"
        elif self.type == T_PRE_REVPROP_CHANGE:
            type = "pre-revprop-change"
        elif self.type == T_POST_REVPROP_CHANGE:
            type = "post-revprop-change"
        elif self.type == T_PRE_LOCK:
            type = "pre-lock"
        elif self.type == T_POST_LOCK:
            type = "post-lock"
        elif self.type == T_PRE_UNLOCK:
            type = "pre-unlock"
        elif self.type == T_POST_UNLOCK:
            type = "post-unlock"
        return type

    def install_config_form(self):
        """
        This method will be called to build setup configuration form.
        If this plugin needs parameters, provides form fields here.
        Any html and javascript are welcome.
        
        If plugin does not need further configuration, simply return null str.
        """
        return ""
        
    def show_install_config_form(self):
        """
        This method will be called to build setup configuration form.
        If this plugin needs parameters, provides form fields here.
        Any html and javascript are welcome.
        
        return output reSTified html.
        """
        header = """
**%(id)s**

- %(t_name)s: %(name)s
- %(t_type)s: %(type)s

**%(t_desc)s**

%(desc)s

%(detail)s
""" % {
                't_name': _('Name'),
                't_type': _('Type'),
                't_desc': _('Description'),
                'id':self.id.rsplit('.',1)[-1],
                'name': self.name,
                'type': self.get_type(),
                'desc': self.description,
                'detail': self.detail,
                }

        header = reSTify(header)
        form = self.install_config_form() or ""
        return header + form
        
    def enabled(self):
        """
        Return True, if this plugin has been setup.
        Simply call 'has_config()'.
        """
        raise Exception, _("Plugin not fully implemented.")
            
    def uninstall(self):
        """
        Uninstall hooks-plugin from repository.
        Simply call 'unset_config()' and 'save()'.
        """
        raise Exception, _("Plugin not fully implemented.")
    
    def install(self, params=None):
        """
        Install hooks-plugin from repository.
        Simply call 'set_config()' and 'save()'.
        
        Form fields in setup_config() will pass as params.
        """
        raise Exception, _("Plugin not fully implemented.")
    
    
modules = getPackageModules(__file__)



"""Pylons environment configuration"""
import os

from mako.lookup import TemplateLookup
from pylons.configuration import PylonsConfig
from pylons.error import handle_mako_error

import pysvnmanager.lib.app_globals as app_globals
import pysvnmanager.lib.helpers
from pysvnmanager.config.routing import make_map

def load_environment(global_conf, app_conf):
    """Configure the Pylons environment via the ``pylons.config``
    object
    """
    config = PylonsConfig()
    
    # Pylons paths
    root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    paths = dict(root=root,
                 controllers=os.path.join(root, 'controllers'),
                 static_files=os.path.join(root, 'public'),
                 templates=[os.path.join(root, 'templates')])

    # Initialize config with the basic options
    config.init_app(global_conf, app_conf, package='pysvnmanager', paths=paths)

    config['routes.map'] = make_map(config)
    config['pylons.app_globals'] = app_globals.Globals(config)
    config['pylons.h'] = pysvnmanager.lib.helpers
    
    # Setup cache object as early as possible
    import pylons
    pylons.cache._push_object(config['pylons.app_globals'].cache)
    

    # Create the Mako TemplateLookup, with the default auto-escaping
    config['pylons.app_globals'].mako_lookup = TemplateLookup(
        directories=paths['templates'],
        error_handler=handle_mako_error,
        module_directory=os.path.join(app_conf['cache_dir'], 'templates'),
        input_encoding='utf-8', default_filters=['escape'],
        imports=['from webhelpers.html import escape'])

    # CONFIGURATION OPTIONS HERE (note: all config options will override
    # any Pylons config options)
    
    # By default, the tmpl_context (a.k.a 'c'), is no longer a AttribSafeContextObj.
    # This means accessing attributes that don't exist will raise an AttributeError.
    # To use the attribute-safe tmpl_context, add this line to the config/environment.py:
    config['pylons.strict_tmpl_context'] = False

    return config



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Validate commit log:
--------------------
* Check the length of commit log;
* Check certain contents against Regular Expression;

Usage:
------
%(program)s [options...] repos txn

Options:
--------
  -s, --size:
      The minimal strlen of log message.

  -p, --permit:
      Pattern for log message should match against.
      Multiple patterns can be given if provide multiple -p,
      if one of the pattern is matched, validation passed.

  -P, --prohibit:
      Pattern that log message should *not* match against.
      Multiple patterns can be given if provide multiple -p,
      if one of the pattern is matched, validation failed.


by Jiang Xin<WorldHello.net.AT.gmail.com>
"""

__revision__ = '$Id: commit_log_check.py 1563 2006-07-04 05:37:20Z jiangxin $'

import sys, os, re
import getopt

if os.name == 'nt':
    SVNLOOK = 'C:/Apps/Subversion/bin/svnlook.exe'
else:
    SVNLOOK = '/usr/bin/svnlook'

os.environ['LANG'] = os.environ['LC_ALL'] = 'zh_CN.utf8'
program = sys.argv[0]

def usage(code=0, msg=''):
    if code:
        fd = sys.stderr
    else:
        fd = sys.stdout
    print >> fd, __doc__ % { 'program':program }
    if msg:
        print >> fd, msg
    return code

def main(argv=None):
    """main entry point"""

    if argv is None:
        argv = sys.argv[1:]

    opt_size = 4
    # r'(issue\s*[#]?[0-9]+)|(new.*:)|(bugfix:)',
    opt_permit_pattern = []
    opt_prohibit_pattern = []
    log_msg = ''

    try:
        opts, args = getopt.getopt( 
                argv, "hs:p:m:P:", 
                ["help", "size=", "permit=", "prohibit=", "message="])
    except getopt.error, msg:
        return usage(1, msg)

    for opt, arg in opts:
        if opt in ("-h", "--help"):
            return usage()
        elif opt in ("-s", "--size"):
            opt_size = int(arg)
        elif opt in ("-p", "--permit"):
            if arg and not isinstance(arg, unicode):
                arg = unicode(arg, 'utf-8')
            if arg:
                opt_permit_pattern.append(arg)
        elif opt in ("-P", "--prohibit"):
            if arg:
                if not isinstance(arg, unicode):
                    arg = unicode(arg, 'utf-8')
                opt_prohibit_pattern.append(arg)
        elif opt in ("-m", "--message"):
            if arg:
                if not isinstance(arg, unicode):
                    arg = unicode(arg, 'utf-8')
                log_msg = arg
        else:
            return usage(1)

    if len(args) != 2:
        return usage(1)

    repos = args[0]
    txn   = args[1]

    if not log_msg:
        log_cmd = '%s log -t "%s" "%s"' % (SVNLOOK, txn, repos)
        log_msg = os.popen(log_cmd, 'r').read().rstrip('\n')

    # Check the length of commit log
    check_strlen(log_msg, opt_size)
    
    # Check certain contents against Regular Expression
    check_pattern(log_msg, opt_permit_pattern, opt_prohibit_pattern)


def check_strlen(log_msg, size):
    """
    Check length of log_msg, not less then size
    """
    log_msg = log_msg.strip()
    log_length = len(log_msg)

    if log_length > 0:
        char  = log_msg[0]
        char2 = log_msg[-1]        
        idx = 1
        while idx < len(log_msg):
            if char == -1 and char2 == -1 and log_length <= 0:
                break

            if (char == log_msg[idx]) and (char != -1):
                log_length = log_length - 1
                char = log_msg[idx]
            else:
                char = -1

            if (char2 == log_msg[-idx]) and (char2 != -1):
                log_length = log_length - 1
                char2 = log_msg[-idx]
            else:
                char2 = -1

            idx = idx + 1
    
    if log_length < size:
        error_msg = u"COMMIT_LOG_TOO_SIMPLE : 提交说明至少应包含 %d 个字符, 或者太简单了。\n" % size
        sys.stderr.write (error_msg.encode('utf-8'))
        sys.exit(1)


def check_pattern(log_msg, permit=None, prohibit=None):
    """
    Check log_msg against patterns
    """

    if permit:
        matched = False
        for pat in permit:
            if not pat: # blank pattern
                matched = True
                break
            elif re.compile(pat, re.I).search(log_msg):
                matched = True
                break
        if not matched:
            error_msg = u"PREDEFINED_PATTERN_NOT_IN_COMMIT_LOG: 无法在提交说明中匹配表达式: \n%s。\n" % ',\n'.join(permit)
            sys.stderr.write (error_msg.encode('utf-8'))
            sys.exit(1)

    if prohibit:
        matched = False
        for pat in prohibit:
            if pat and re.compile(pat, re.I).search(log_msg):
                matched = True
                break
        if matched:
            error_msg = u"BAD_COMMIT_LOG : 不允许在log中出现类似表达式: \n%s。\n" % pat
            sys.stderr.write (error_msg.encode('utf-8'))
            sys.exit(1)

if __name__ == '__main__':
    main()

# vim: ft=python ts=4 sw=4 et



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2008 OpenSourceXpress Ltd. (http://www.ossxp.com)
# Author: Jiang Xin
# Contact: http://www.ossxp.com
#          http://www.worldhello.net
#          http://moinmo.in/JiangXin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

from pysvnmanager.hooks.plugins import *
from pysvnmanager.hooks.plugins import _
import webhelpers

class EmailNotify(PluginBase):

    # Brief name for this plugin.
    name = _("Send email notify for commit event")
    
    # Both description and detail are reStructuredText format. 
    # Reference about reStructuredText: http://docutils.sourceforge.net/docs/user/rst/quickref.html

    # Short description for this plugin.
    description = _("Send a notification email describing either a commit or "
                    "a revprop-change action on a Subversion repository.")
    
    # Long description for this plugin.
    detail = _("""
You must provide proper options to commit-email.pl using the
configuration form for this plugin.

You can simply just provide the email_addr as the options.

  [options] email_addr [email_addr ...]

But to be more versitile, you can setup a path-based email 
notifier.

  [-m regex1] [options] [email_addr ...]
  [-m regex2] [options] [email_addr ...] 
  ...

Options:

-m regex              Regular expression to match committed path
--from email_address  Email address for 'From:' (overrides -h)
-r email_address      Email address for 'Reply-To:
-s subject_prefix     Subject line prefix
--diff n              Do not include diff in message (default: y)
""")
    
    # Hooks-plugin type: T_START_COMMIT, ..., T_POST_UNLOCK
    type = T_POST_COMMIT
    
    # Plugin config option/value in config ini file.
    # email_notify_enable = yes|no
    # email_notify_config = -m "." --diff y --from <from@addr> -r <reply@addr> -s "[prefix]" to@addr
    key_switch = "email_notify_enable"
    key_config = "email_notify_config"
    
    section = 'email'
    
    def enabled(self):
        """
        Return True, if this plugin has been installed.
        Simply call 'has_config()'.
        """
        return self.has_config(self.key_switch) and self.has_config(self.key_config)
    
    def install_info(self):
        """
        Show configurations if plugin is already installed.
        
        return reStructuredText.
        reST reference: http://docutils.sourceforge.net/docs/user/rst/quickref.html
        """
        result = self.description
        if self.enabled():
            result += "\n\n"
            result += "**" + _("Current configuration") + "**\n\n"
            if self.get_config(self.key_switch) == "yes":
                result += "- " + _("Email notify enabled.")
            else:
                result += "- " + _("Email notify disabled.")
            result += "\n"
            result += "- " + _("Parameters: ") + "``" + self.get_config(self.key_config) + "``\n"

        return result
    
    def install_config_form(self):
        """
        This method will be called to build setup configuration form.
        If this plugin needs parameters, provides form fields here.
        Any html and javascript are welcome.
        """
        if self.get_config(self.key_switch)=="no":
            enable_checked  = ""
            disable_checked = "checked"
        else:
            enable_checked  = "checked"
            disable_checked = ""

        result = ""
        result += "<p><strong>%s</strong></p>" % _("Fill this form")
        result += "<blockquote>"
        result += "<dl>"
        result += "\n<dt>"
        result += _("Enable email notify.")
        result += "\n<dd>"
        result += "<input type='radio' name='switch' value='yes' " + \
                enable_checked  + ">" + _("Enable") + "&nbsp;"
        result += "<input type='radio' name='switch' value='no' " + \
                disable_checked + ">" + _("Disable") + "<br>"
        result += "\n<dt>"
        result += _("Input email notify configurations: ")
        result += "\n<dd>"
        result += "<textarea name='config' rows='5' cols='40'>"
        result += webhelpers.util.html_escape(self.get_config(self.key_config))
        result += "</textarea>"
        result += "\n</dl>"
        result += "</blockquote>"
        return result
        
    def uninstall(self):
        """
        Uninstall hooks-plugin from repository.
        Simply call 'unset_config()' and 'save()'.
        """
        self.unset_config(self.key_config)
        self.unset_config(self.key_switch)
        self.save()
    
    def install(self, params=None):
        """
        Install hooks-plugin from repository.
        Simply call 'set_config()' and 'save()'.
        
        Form fields in setup_config() will pass as params.
        """
        switch = params.get('switch', 'yes')
        if switch != 'yes':
            switch = 'no'
        config = params.get('config')
        if not config:
            raise Exception, _("Wrong configuration.")
        self.set_config(self.key_switch, switch)
        self.set_config(self.key_config, config)
        self.save()
        
def execute(repospath=""):
    """
    Generate and return a hooks plugin object

    @param request: repos full path
    @rtype: Plugin
    @return: Plugin object
    """
    return EmailNotify(repospath)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
import os
import os.path
from svn import repos, fs, core

def main(pool, repos_dir, path):
    # Construct a ChangeCollector to fetch our changes.
    fs_ptr = repos.svn_repos_fs(repos.svn_repos_open(repos_dir, pool))
    youngest_rev = fs.svn_fs_youngest_rev(fs_ptr, pool)
    root = fs.svn_fs_revision_root(fs_ptr, youngest_rev, pool)
    if not fs.svn_fs_node_prop(root, path, core.SVN_PROP_NEEDS_LOCK, pool):
        sys.stderr.write(
"""Locking of path '%s' prohibited by repository policy (must have
%s property set)
""" % (path, core.SVN_PROP_NEEDS_LOCK))
        return 1
    return 0


def _usage_and_exit():
    sys.stderr.write("""
Usage: %s REPOS-DIR PATH

This script, intended for use as a Subversion pre-lock hook, verifies that
the PATH that USER is attempting to lock has the %s property
set on it, returning success iff it does.
""" % (os.path.basename(sys.argv[0]), core.SVN_PROP_NEEDS_LOCK))
    sys.exit(1)

    
if __name__ == '__main__':
    if len(sys.argv) < 3:
        _usage_and_exit()
    sys.exit(core.run_app(main, sys.argv[1], sys.argv[2]))
  



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008 OpenSourceXpress Ltd. (http://www.ossxp.com)
# Author: Jiang Xin
# Contact: http://www.ossxp.com
#          http://www.worldhello.net
#          http://moinmo.in/JiangXin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

"""Setup the pySvnManager application"""
import logging

import pylons.test

from pysvnmanager.config.environment import load_environment

from paste.deploy import appconfig
from shutil import copyfile
import os
from pkg_resources import resource_filename

from pysvnmanager.config.environment import load_environment

log = logging.getLogger(__name__)

def setup_app(command, conf, vars):
    """Place any commands to setup pysvnmanager here"""
    # Don't reload the app if it was loaded under the testing environment
    if not pylons.test.pylonsapp:
        load_environment(conf.global_conf, conf.local_conf)
    else:
        # Hack pylons: config['here'] is used in many places, so setup here.
        from pylons import config
        wsgiapp = pylons.test.pylonsapp
        config['here'] = wsgiapp.config.get('here')

    here = conf['here']

    if not os.path.exists(here+'/config'):
        os.mkdir(here+'/config')
    if not os.path.exists(here+'/config/RCS'):
        os.mkdir(here+'/config/RCS')
    if not os.path.exists(here+'/svnroot'):
        os.mkdir(here+'/svnroot')
    filelist = ['svn.access', 'svn.passwd', 'localconfig.py']
    for f in filelist:
        src  = resource_filename('pysvnmanager', 'config/' + f+'.in')
        dest = here+'/config/' + f
        if os.path.exists(dest):
            log.warning("Warning: %s already exist, ignored." % f)
        else:
            copyfile(src, dest)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2008 OpenSourceXpress Ltd. (http://www.ossxp.com)
# Author: Jiang Xin
# Contact: http://www.ossxp.com
#          http://www.worldhello.net
#          http://moinmo.in/JiangXin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

import os

svn_hooks_init_dict = { '1.4': 'hook1.4',
                        '1.5': 'hook1.5',
                        'default': 'hook1.5',
                      }

svn_hooks_init_base = os.path.dirname(os.path.abspath(__file__))
